import impmagic

class StructureCorrect:
    def __init__(self, projectfolder, exclude='dist'):
        self.projectfolder = projectfolder
        self.exclude = exclude

    @impmagic.loader( 
        {'module':'os'}, 
        {'module':'os.path', 'submodule': ['join']}
    )
    def init(self):            
        self.initial_files_dirs = set(join(dp, f) for dp, dn, filenames in os.walk(self.projectfolder) for f in filenames + dn)

    @impmagic.loader(
        {'module':'shutil'}, 
        {'module':'os'}, 
        {'module':'os.path', 'submodule': ['join', 'isdir', 'exists', 'abspath']}
    )
    def purge(self):
        final_files_dirs = set(join(dp, f) for dp, dn, filenames in os.walk(self.projectfolder) for f in filenames + dn)
        self.result = {'dirs':[], 'files':[]}
        exclude = join(abspath('.'), self.exclude)
        final_files_dirs = sorted(final_files_dirs)
        for file in final_files_dirs:
            if file not in self.initial_files_dirs:
                if not file.startswith(exclude):
                    print(file)
                    find = False
                    for element in self.result['dirs']:
                        if file.startswith(element):
                            find = True

                    if not find:
                        if isdir(file):
                            if not file.endswith(self.exclude):
                                shutil.rmtree(file)
                            else:
                                self.result['dirs'].append(file)
                        else:
                            if exists(file):
                                os.remove(file)
                                self.result['files'].append(file)


#Récupère une liste de fichiers et retourne une liste de chemin
@impmagic.loader(
    {'module':'glob', 'submodule': ['glob']}
)
def list_to_content(content_backup):
    result = []
    for element in content_backup:
        if len(element):
            if isinstance(element, str):
                if exists(element):
                    if isfile(element):
                        result.append(element)
                    elif isdir(element):
                        folder = glob(element+"/*")+glob(element+"/.*")
                        if len(folder):
                            res = list_to_content(folder)
                            if len(res):
                                result+=res
                        else:
                            #pour inclure les dossiers vide
                            result.append(element)
                else:
                    logs(f"{element} introuvable", "warning")
            else:
                res = list_to_content(element)
                if len(res):
                    result+=res
    return result



@impmagic.loader(
    {'module':'template.toml_format', 'submodule': ['DEFAULT_TOML']},
    {'module':'app.display', 'submodule': ['logs']},
    {'module':'os.path', 'submodule': ['exists', 'isfile', 'isdir']}
)
def list_projectfile(tfile):
    project_dir = tfile.get_key(DEFAULT_TOML['projectname']['name'],DEFAULT_TOML['projectname']['section'])
    include_files = tfile.get_key(DEFAULT_TOML['include_files']['name'],DEFAULT_TOML['include_files']['section'])
    
    license_file = tfile.get_key(DEFAULT_TOML['license_file']['name'],DEFAULT_TOML['license_file']['section'])
    readme = tfile.get_key(DEFAULT_TOML['readme']['name'],DEFAULT_TOML['readme']['section'])
    changelog = tfile.get_key(DEFAULT_TOML['changelog']['name'],DEFAULT_TOML['changelog']['section'])

    content_backup = ['nexus.toml', '.package_alias', project_dir, include_files, license_file, readme, changelog]

    content = list_to_content(content_backup)

    return content


#Parse les fichiers du projet pour en faire un dictionnaire et compresse le contenu
@impmagic.loader(
    {'module':'brotli'},
    {'module':'hashlib'},
    {'module':'glob', 'submodule': ['glob']},
    {'module':'os.path', 'submodule': ['exists', 'isfile', 'isdir', 'basename']}
)
def dump_project(projectfolder):
    result = {}

    for file in glob(f"{projectfolder}/*"):
        if isdir(file):
            if basename(file)!="__pycache__":
                result[basename(file)] = dump_project(file)
        else:
            with open(file, 'rb') as f:
                content = f.read()

                result[basename(file)] = {
                    'hash': hashlib.sha256(content).hexdigest(),
                    'content': brotli.compress(content, quality=11).hex()
                }
    return result


@impmagic.loader(
    {'module':'hashlib'},
    {'module':'brotli'},
    {'module':'app.display', 'submodule': ['logs']},
    {'module':'os.path', 'submodule': ['exists', 'join']},
    {'module':'structure.check', 'submodule': ['get_root_project']}
)
def load_file(specific_file, deps, dest=None):
    if specific_file in deps:
        logs(f"Restauration du {specific_file}")

        if specific_file=='project_file':
            deps[specific_file]['name'] = "nexus.toml"

        if 'hash' in deps[specific_file]:
            if 'name' not in deps[specific_file]:
                logs(f"Le snap ne contient pas le nom du fichier {specific_file}", "critical")
                return

            if dest:
                sp_file = join(dest, deps[specific_file]['name'])
            else:
                sp_file = join(get_root_project(), deps[specific_file]['name'])
            inject = True

            if exists(sp_file):
                with open(sp_file, 'rb') as f:
                    hash_file =  hashlib.sha256(f.read()).hexdigest()
                    if hash_file==deps[specific_file]['hash']:
                        inject=False
                        logs(f"Fichier {specific_file} inchangé")

            if 'content' in deps[specific_file]:
                if inject:
                    with open(sp_file, 'wb') as f:
                        f.write(brotli.decompress(bytes.fromhex(deps[specific_file]['content'])))
            else:
                logs(f"Le snap ne contient pas le contenu du fichier {specific_file}")
        else:
            logs(f"Le snap ne contient pas le hash du fichier {specific_file}")


@impmagic.loader(
    {'module':'hashlib'},
    {'module':'brotli'},
    {'module':'os', 'submodule': ['mkdir']},
    {'module':'app.display', 'submodule': ['logs']},
    {'module':'os.path', 'submodule': ['exists', 'isfile', 'join']},
    {'module':'structure.check', 'submodule': ['get_root_project']}
)
def load_project(project_data, dest_dir):
    if not exists(dest_dir):
        mkdir(dest_dir)

    for namefile, element in project_data.items():
        sp_file = join(dest_dir, namefile)
        if 'hash' in element and 'content' in element:
            inject = True

            if exists(sp_file):
                with open(sp_file, 'rb') as f:
                    hash_file =  hashlib.sha256(f.read()).hexdigest()
                    if hash_file==element['hash']:
                        inject=False
                        logs(f"Fichier {sp_file} inchangé")

            if inject:
                logs(f"Restauration de {sp_file}")
                with open(sp_file, 'wb') as f:
                    f.write(brotli.decompress(bytes.fromhex(element['content'])))
        else:
            load_project(element, sp_file)


def tree_project(project_data):
    result = {}

    for namefile, element in project_data.items():
        if 'hash' in element and 'content' in element:
            result[namefile] = ''
        else:
            result[namefile] = tree_project(element)

    return result


def select_file_from_project(file, project_data):
    inject = {}

    if isinstance(file, str):
        file = file.replace("\\", "/").split("/") 
    
    if len(file):
        element = file.pop(0)

        if element in project_data:
            if isinstance(project_data[element], dict) and 'hash' in project_data[element]:
                inject[element] = project_data[element]
            
            elif isinstance(project_data[element], dict) and 'hash' not in project_data[element]:
                if len(file):
                    result = select_file_from_project(file, project_data[element])
                else:
                    result = project_data[element]

                if result:
                    inject[element] = result

            else:
                return None

    return inject


#Compare le code du projet actuel avec le code du snap
@impmagic.loader(
    {'module':'hashlib'}, 
    {'module':'glob', 'submodule': ['glob']},
    {'module':'os.path', 'submodule': ['isdir', 'join', 'basename']},
    {'module':'app.display', 'submodule': ['logs']}
)
def compare_snap_file(projectfolder, project_data, origin=None):
    if not origin:
        origin = projectfolder.replace("\\", "/")

    for file in glob(f"{projectfolder}/*"):
        if basename(file)!="__pycache__":
            if basename(file) in project_data:
                if isdir(file):
                    compare_snap_file(file, project_data[basename(file)], origin)
                else:
                    with open(file, 'rb') as f:
                        hash_file = hashlib.sha256(f.read()).hexdigest()

                    filename = file.replace("\\","/").replace(f'{origin}/', '')
                    
                    if project_data[basename(file)]['hash']==hash_file:
                        logs(f"Fichier {filename} inchangé")
                    else:
                        logs(f"Fichier {filename} modifié", "warning")
            else:
                logs(f"{join(projectfolder, file)} nouvellement créé", "warning")


import impmagic

#Convertis un str en liste, contrôle les regex associés pour valider le format et sortie en dict
@impmagic.loader(
    {'module':'template.toml_format', 'submodule':['DEFAULT_TOML']},
    {'module':'re', 'submodule':['match']}
)
def check_list(value, name):
    if isinstance(value, str):
        value = value.split(",")

    result = {}
    if 'regex' in DEFAULT_TOML[name]:
        for element in value:
            if match(DEFAULT_TOML[name]["regex"], element):
                result.append(element)

        return result
    else:
        return value

#Formate une liste de dépendance en dictionnaire
@impmagic.loader(
    {'module':'template.regex', 'submodule':['dependencies']},
    {'module':'re', 'submodule':['match']}
)
def check_dependency_format(value):
    if isinstance(value, str):
        value = value.split(",")
    elif isinstance(value, dict):
        return value

    result = {}
    for package in value:
        matchs = match(dependencies, package)
        if matchs!=None:
            package_name = matchs.group(1)
            version = matchs.group(2)
            if version==None:
                version = "N.A"

            result[package_name]=version

    return result


import impmagic

class Find_modules:
    @impmagic.loader(
        {'module':'ast'},
        {'module': 'os.path', 'submodule': ['exists']},
        {'module': 'app.display', 'submodule': ['logs']}
    )
    def __init__(self, file, submodule=False):
        if exists(file):
            with open(file) as f:
                self.tree = ast.parse(f.read())
        else:
            logs(f"Fichier {file} introuvable", "critical")
            exit()
        self.result = []
        self.submodule=submodule

    @impmagic.loader(
        {'module':'ast'}
    )
    def get_variable(self, name):
        for node in ast.walk(self.tree):
            if isinstance(node, ast.Assign):
                if isinstance(node.targets[0], ast.Name) and node.targets[0].id == name:
                    return node.value
            if isinstance(node, ast.For):
                target_node = node.target
                iter_node = node.iter
                if isinstance(target_node, ast.Name) and target_node.id == name:
                    if isinstance(iter_node, ast.Name):
                        data = self.get_variable(iter_node.id)
                        if isinstance(data, ast.Constant):
                            return data.value
                        else:
                            return self.get_value(data)

    @impmagic.loader(
        {'module':'ast'}
    )
    def get_value(self, data, result=None):
        try:
            if isinstance(data, ast.Constant):
                return data.value

            elif isinstance(data, ast.Tuple):
                result = ()
                for element in data.elts:
                    if isinstance(element, ast.Constant):
                        result = result + (element,)
                    else:
                        result = result + (self.get_value(element), )

            elif isinstance(data, ast.Dict):
                result = {}
                for i, value in enumerate(data.values):
                    if isinstance(value, ast.Constant):
                        result[data.keys[i].value] = value.value
                    else:
                        value = self.get_value(value)
                        result[data.keys[i].value] = value

            elif isinstance(data, ast.List):
                result = []
                for element in data.elts:
                    if isinstance(element, ast.Constant):
                        result.append(element.value)
                    else:
                        result.append(self.get_value(element))

            elif isinstance(data, ast.Subscript):
                key = self.get_value(data.slice)
                res_var = self.get_variable(data.value.id)
                
                result = []
                for el in res_var:
                    if el!=None and isinstance(el, dict):
                        result.append(el[key])
                    else:
                        result = res_var

            elif isinstance(data, ast.Name):
                res_var = self.get_variable(data.id)
                if isinstance(res_var, list) or isinstance(res_var, tuple):
                    result = []
                    for element in res_var:
                        if isinstance(element, str):
                            result.append(element)
                        else:
                            result.append(self.get_value(element))
                elif isinstance(res_var, str):
                    result = res_var
                else:
                    result = self.get_value(res_var)
            return result
        except:
            return None

    def add_module(self, name):
        if not name.startswith('__'):
            if not self.submodule and '.' in name:
                name = name.split(".")[0]

            if name not in self.result:
                self.result.append(name)

    @impmagic.loader(
        {'module':'ast'}
    )
    def search(self):
        for node in ast.walk(self.tree):
            if isinstance(node, ast.Call):
                if isinstance(node.func, ast.Name):
                    if hasattr(node.func, 'id') and node.func.id=='__import__' or node.func.id=="import_module":
                        res = self.get_value(node.args[0])
                        if isinstance(res, list):
                            for element in res:
                                self.add_module(element)
                        else:
                            if res!=None:
                                self.add_module(res)
                if isinstance(node.func, ast.Attribute):
                    if hasattr(node.func.value, 'id') and  (node.func.value.id=="importlib" and node.func.attr=="import_module"):
                        self.add_module(self.get_value(node.args[0]))
                    elif hasattr(node.func.value, 'id') and  (node.func.value.id=="impmagic" and node.func.attr=="loader"):
                        for element in node.args:
                            data = self.get_value(element)
                            if isinstance(data, dict) and 'module' in data:
                                if 'submodule' in data and isinstance(data['submodule'], list):
                                    for submodule in data['submodule']:
                                        self.add_module(data['module']+"."+submodule)
                                else:
                                    self.add_module(data['module'])

            elif isinstance(node, ast.Import):
                for name in node.names:
                    self.add_module(name.name)
            elif isinstance(node, ast.ImportFrom):
                self.add_module(node.module)

@impmagic.loader(
    {'module':'os.path', 'submodule': ['exists', 'isdir', 'isfile']},
    {'module':'glob', 'submodule': ['glob']},
    {'module':'package.check', 'submodule': ['is_native_module']},
)
def get_modules_from_project(files, recursive=True):
    data = {'interne':[], 'dependency':[], 'externe':[]}
    analysed = []
    search = True
    while search:
        new = []
        for file in files:
            if file not in analysed:
                analysed.append(file)
            a = Find_modules(file, submodule=False)
            a.search()
            for mod in a.result:
                if exists(mod) and isdir(mod):
                    if mod not in data['interne']:
                        data['interne'].append(mod)
                    if recursive:
                        for mod_file in glob(mod + '/**/*.py', recursive=True):
                            if mod_file not in analysed:
                                new.append(mod_file)

                elif exists(mod+".py") and isfile(mod+".py"):
                    if mod+".py" not in analysed and recursive:
                        new.append(mod+".py")
                    if mod not in data['interne']:
                        data['interne'].append(mod)
                elif exists(mod+".pyw") and isfile(mod+".pyw"):
                    if mod+".pyw" not in analysed and recursive:
                        new.append(mod+".pyw")
                    if mod not in data['interne']:
                        data['interne'].append(mod)
                else:
                    if is_native_module(mod):
                        if mod not in data['externe']:
                            data['externe'].append(mod)
                    else:
                        if mod not in data['dependency']:
                            data['dependency'].append(mod)

        if len(new)==0:
            search = False
        
        files = new

    return data


@impmagic.loader(
    {'module':'app.display', 'submodule': ['prompt']},
    {'module':'package.package', 'submodule': ['in_repo', 'get_alias', 'set_alias']}
)
#Liste les packages dont les noms d'install sont différents et sans alias
def check_realname(package, init=False):
    alias = get_alias(init)
    insert_alias = alias.copy()

    if package in alias:
        fnd = alias[package]
        if isinstance(fnd, str):
            return fnd
        else:
            data = {}
            if 'windows' in fnd:
                if in_repo(fnd['windows']):
                    data['windows'] = fnd['windows']
            if 'linux' in fnd:
                if in_repo(fnd['linux']):
                    data['linux'] = fnd['linux']

            if len(data):
                return data
    else:
        packagename = package
        while not in_repo(package):
            package = prompt(packagename)
            if package=="{exclude}":
                package = None
                break

        if package and packagename!=package:
            insert_alias[packagename] = package

    if alias!=insert_alias:
        set_alias(insert_alias)

    return package


import __main__
import impmagic

"""
import xmlrpc.client
client = xmlrpc.client.ServerProxy('https://pypi.org/pypi')
client.package_releases('py-cpuinfo')
client.release_data('py-cpuinfo', '0.1.3')
client.list_packages()
"""

#A REFAIRE
def build_cache():
    import xmlrpc.client
    client = xmlrpc.client.ServerProxy(f'{__main__.nxs.metadata_source}/pypi')
    
    print("Récupération de la liste")
    packages = client.list_packages()
    maxd = len(packages)

    print("Construction")
    i=1
    for package in packages:
        print(f"{i}/{maxd} - {package}")
        get_package(package)
        i+=1

#Récupération les informations d'un package
@impmagic.loader(
    {'module':'json'},
    {'module':'sys_nxs.connect', 'submodule':['get']},
    {'module':'cache.sqlite', 'submodule':['Cache']},
    {'module': 'template', 'submodule': ['default_conf']},
    {'module': 'sys_nxs.host', 'submodule': ['path_rep']},
    {'module':'os.path', 'submodule': ['split', 'expanduser', 'join']}
)
def get_package_information(package_name, version):
    cachefile = join(expanduser(__main__.nxs.conf.load(val='cache-dir', section='',default=default_conf.cache_dir)).replace(path_rep[1], path_rep[0]), 'repo_cache.db')

    dbcache = Cache(cachefile)

    result = dbcache.search(package_name, version)

    if len(result):
        return result[0]

    else:
        url = f"{__main__.nxs.metadata_source}/pypi/{package_name}/{version}/json"

        response = get(url)
        if response.status_code == 200:
            insert_data = {}
            data = response.json()
            info = data['info']

            informations = ['name', 'version', 'summary', 'author', 'maintainer', 'license', 'home_page', 'requires_python', 'requires_dist', 'vulnerabilities']

            for inf in informations:
                if inf in info and info[inf]!=None and len(info[inf])>0:
                    if isinstance(info[inf], list):
                        insert_data[inf] = json.dumps(info[inf])
                    else:
                        insert_data[inf] = info[inf]
                else:
                    insert_data[inf] = ""

            dbcache.add(insert_data.copy())

            return insert_data


#Récupération des informations d'un package
@impmagic.loader(
    {'module':'concurrent.futures', 'as': 'worker'},
    {'module':'sys_nxs.connect', 'submodule':['get']},
    {'module': 'template', 'submodule': ['default_conf']}
)
def get_package(package_name, version=None):
    if version!=None:
        return get_package_information(package_name, version)
    else:
        releases = get_versionlist_package(package_name)
        
        result = []
        if __main__.nxs.conf.load(val='threading', section='',default=default_conf.threading):
            with worker.ThreadPoolExecutor(max_workers=__main__.nxs.threadmax) as executor:
                futures = [executor.submit(get_package_information, package_name, release) for release in releases]

                worker.wait(futures)

                result = [future.result() for future in futures]


        else:
            for release in releases:
                result.append(get_package_information(package_name, release))

        return result


#Récupération des la liste des requirements
@impmagic.loader(
    {'module':'concurrent.futures', 'as': 'worker'},
    {'module': 'template', 'submodule': ['default_conf']}
)
def get_all_requirement(used, i=0):
    if i>__main__.nxs.conf.load(val='project.deep_search', section='',default=default_conf.project_deep_search):
        return used

    if __main__.nxs.conf.load(val='threading', section='',default=default_conf.threading):
        with worker.ThreadPoolExecutor(max_workers=__main__.nxs.threadmax) as executor:
            futures = [executor.submit(run_all_requirement, used_package, used, i) for used_package in used.copy()]

            worker.wait(futures)

            for future in futures:
                used = list(set(used).union(future.result()))
    else:
        for used_package in used.copy():
            requirement_package = run_all_requirement(used_package, used, i)
            used = list(set(used).union(requirement_package))

    return used


def run_all_requirement(used_package, used, i):
    v = get_versionlist_package(used_package)
    if len(v):
        try:
            requirements = get_requirement(used_package, v[-1])
            if len(requirements):
                # Créer une liste avec les éléments de liste2 qui ne sont pas dans liste1
                resultat = [x for x in list(requirements.keys()) if x not in used]

                used = list(set(used).union(list(requirements.keys())))

                req = get_all_requirement(resultat, i+1)
                if len(req):
                    used = list(set(used).union(req))
        except:
            pass
    return used


#Récupération de la liste des dépendances d'un package
@impmagic.loader(
    {'module':'os'},
    {'module':'re'},
    {'module':'json'},
    {'module':'sys_nxs.connect', 'submodule':['get']},
    {'module':'template', 'submodule':['regex']},
    {'module':'cache.direct', 'submodule':['DirectCache']}
)
def get_requirement(package_name, version, test=False):
    cache = DirectCache()
    data = cache.get("package_info", package_name)

    if data==None:
        data = get_package_information(package_name, version)
        if data==None:
            return {}

        cache.set("package_info", package_name, data)
    dependencies = data['requires_dist']
    if dependencies==None:
        return {}
    else:
        package_compiled = re.compile(regex.package_regex)
        python_version_compiled = re.compile(regex.python_version_regex)
        sys_platform_compiled = re.compile(regex.sys_platform_regex)
        extra_compiled = re.compile(regex.extra_regex)
        operator_regex = re.compile(regex.segment_version)

        result = {}

        try:
            dependencies = json.loads(dependencies)
        except:
            pass

        for deps in list(dependencies):
            package = package_compiled.search(deps)
            if package!=None:
                name = package.group('name')
                version = package.group('version')
                #print(f"Name: {package.group('name')}\nVersion: {package.group('version')}")
                
                pyver = python_version_compiled.search(deps)
                if pyver!=None and hasattr(__main__.nxs, 'py_version'):
                    pyversion = pyver.group('python_version')
                    pyversion = pyversion.replace("'","").replace(" ","")
                    match = operator_regex.match(pyversion)
                    if match:
                        operator = match.group('operator')
                        target_major = match.group('major')
                        target_minor = match.group('minor')
                        target_patch = match.group('patch')

                        target = target_major
                        if target_minor!=None:
                            target+= "."+target_minor
                        if target_patch!=None:
                            target+= "."+target_patch

                        if not compare_version(__main__.nxs.py_version, operator, target):
                            continue

                    #print(f"Pyversion: {pyversion}")

                platf = sys_platform_compiled.search(deps)
                if platf!=None:
                    if platf.group('platform').lower().startswith('win') and os.name!='nt':
                        continue
                    #print(f"platform: {platf.group('platform')}")

                extra = extra_compiled.search(deps)
                if extra!=None:
                    if extra.group('extra')=="test" and test==False:
                        continue
                    #print(f"extra: {extra.group('extra')}")

                if name!=None:
                    result[name] = {}
                    if version!=None:
                        result[name]['version'] = version
                    else:
                        result[name]['version'] = 'N.A'

        #return dependencies
        #print(result)
        return result
    return {}

#Vérifie la compatibilité du module avec les autres modules a installer
@impmagic.loader(
    {'module':'template', 'submodule':['regex']},
    {'module':'re'}
)
def get_conflit_package(name, dictionary, element=None):
    if element is None or 'compatibility' not in element or not element['compatibility']:
        return dictionary

    version = element['compatibility'][-1]
    operator_regex = re.compile(regex.segment_version)
    
    # Recherche des dépendances du package
    requirements = get_requirement(name, version)

    for package_name, requirement in requirements.items():
        if requirement['version'] == 'N.A':
            compatible_version = get_compatible_versions(package_name, ">=0.0.0")
        else:
            number_regex = re.compile(regex.dissociate_version)
            compatible_version = None

            for match in number_regex.finditer(requirement['version']):
                dep_version = match.group()
                match = operator_regex.match(dep_version)

                possible = get_compatible_versions(package_name, dep_version)
                if compatible_version is None:
                    compatible_version = set(possible)
                else:
                    compatible_version.intersection_update(possible)

            if package_name in dictionary:
                conflit = all(ver not in compatible_version for ver in dictionary[package_name]['compatibility'])

                if conflit:
                    if 'conflicts' not in dictionary[name]:
                        dictionary[name]['conflicts'] = []
                    dictionary[name]['conflicts'].append(f"{package_name} ({requirement})")
                    if len(element['compatibility']) > 1:
                        element['compatibility'].pop()
                        dictionary[name]['compatibility'] = element['compatibility']

    return dictionary




#Récupère un dictionnaire (du toml) et une dictionnaire avec les versions compatibles
@impmagic.loader(
    {'module':'template', 'submodule':['regex']},
    {'module':'app.display', 'submodule':['logs']},
    {'module': 'template', 'submodule': ['default_conf']},
    {'module':'package.check', 'submodule':['is_native_module']},
    {'module':'re'}
)
def dependency_to_list(dictionary, force=False, nocheck=False):
    full_log = __main__.nxs.conf.load(val='logs.full', section='',default=default_conf.logs_full)
    result = {}
        #result = []

    logs("Analyse des dépendances")
    """
    {'module':'cache.direct', 'submodule':['DirectCache']}
    cache = DirectCache()
    cache.set("version_package", package_name, result)
    """
    logs("Téléchargement des informations")
    pull_package_informations(dictionary)
    #exit()

    dictionary = construct_dependencies(dictionary)

    if not nocheck:
        #Analyse les dépendances
        version_regex = re.compile(regex.dependencies)
        logs(f"Recherche de conflit")
        for name, element in dictionary.items():
            dictionary = get_conflit_package(name, dictionary, element)


    #Construct la liste des modules à installer avec la version la plus récente compatible
    for element in dictionary:
        if dictionary[element]['compatibility']==None:
            if not is_native_module(element):
                logs(f"Module {element} non disponible", "error")
        else:
            if 'conflicts' in dictionary[element].keys():
                logs(f"Aucune version compatible pour le module {element}\nConflits avec les packages:\n  - {'  - '.join(dictionary[element]['conflicts'])}", "warning")
                if not force:
                    return None
            else:
                if len(dictionary[element]['compatibility'])>0:
                    result[str(element)] = {}
                    #Récupération de la version la plus récente
                    result[str(element)]['version'] = max(dictionary[element]['compatibility'], key=lambda x: [int(i) for i in x.split('.')])
                    result[str(element)]['compatible'] = dictionary[element]['compatibility']
                    #result.append(str(element)+"=="+dictionary[element]['compatibility'].pop())
                else:
                    logs(f"Aucune version compatible pour le module {element}", "warning")

    return result


#Télécharge les informations des packages
@impmagic.loader(
    {'module':'app.display', 'submodule':['logs']},
    {'module':'concurrent.futures', 'as': 'worker'},
    {'module':'os'}
)
def pull_package_informations(dictionary, result=None):
    with worker.ThreadPoolExecutor(max_workers=__main__.nxs.threadmax) as executor:
        futures = []
        for key, value in dictionary.items():
            futures.append(executor.submit(thread_pull, key, value))
        worker.wait(futures)


def thread_pull(key, value):
    if isinstance(value, dict):
        if key=="windows" and os.name=="nt":
            result = thread_pull(value, result)
        elif key=="linux" and os.name!="nt":
            result = thread_pull(value, result)
        elif key!="windows" and key!="linux":
            result = thread_pull(value, result)

    else:
        v = get_versionlist_package(key)
        if len(v):
            requirements = get_requirement(key, v[-1])
            if len(requirements):
                for requirement in requirements:
                    get_versionlist_package(requirement)  



#Construct un dictionnaire contenant chaque package avec une liste de version compatible
@impmagic.loader(
    {'module':'app.display', 'submodule':['logs']},
    {'module':'functools', 'submodule':['reduce']},
    {'module': 'template', 'submodule': ['default_conf']},
    {'module':'os'}
)
def construct_dependencies(dictionary, result=None):
    full_log = __main__.nxs.conf.load(val='logs.full', section='',default=default_conf.logs_full)

    if result==None:
        result = {}

    for key, value in dictionary.items():
        if isinstance(value, dict):
            if key=="windows" and os.name=="nt":
                result = construct_dependencies(value, result)
            elif key=="linux" and os.name!="nt":
                result = construct_dependencies(value, result)
            elif key!="windows" and key!="linux":
                result = construct_dependencies(value, result)

        else:
            """
            if full_log:
                logs(f"Analyse de {key}")
            """
            if value=="N.A":
                compatible_version = get_compatible_versions(key, ">=0.0.0")
            else:
                if ',' in value:
                    compatible_version = []
                    for ver in value.split(","):
                        compatible_version.append(get_compatible_versions(key, ver))

                    common_versions = reduce(set.intersection, [set(versions) for versions in compatible_version])
                    compatible_version = sorted(common_versions, key=lambda x: tuple(map(int, x.split('.'))))
                else:
                    compatible_version = get_compatible_versions(key, value)

            if key in result.keys():
                if result[key]['compatibility']!=None:
                    for ver in result[key]['compatibility']:
                        if ver not in compatible_version:
                            result[key]['compatibility'].remove(ver)
                else:
                    logs(f"Package {key} non compatible", "warning")
            else:
                result[key] = {}
                result[key]['compatibility'] = compatible_version
    return result

#Retourne la liste des versions disponibles d'un package
@impmagic.loader(
    {'module':'sys_nxs.connect', 'submodule':['get']},
    {'module':'cache.direct', 'submodule':['DirectCache']}
)
def get_versionlist_package(package_name):
    cache = DirectCache()
    info = cache.get("version_package", package_name)

    if info!=None:
        return info

    result = []

    headers = {"Accept": "application/vnd.pypi.simple.v1+json"}
    #print("aaaaaaa")
    #import requests
    response = get(f"{__main__.nxs.metadata_source}/simple/{package_name}/", headers=headers)
    #print(info.text)
    #url = f"https://pypi.org/pypi/{package_name}/json"
    #response = get(url)
    if response.status_code == 200:
        data = response.json()
        releases = data["versions"]
        for version in releases:
            result.append(version)

    cache.set("version_package", package_name, result)
    return result


@impmagic.loader(
    {'module':'template', 'submodule':['regex']},
    {'module':'re'}
)
def get_compatible_versions(package, version=None):
    # Récupérer la liste des versions disponibles
    releases = get_versionlist_package(package)

    if len(releases)>0:
        if version==None:
            return releases

        # Liste qui va stocker les versions compatibles
        compatible_versions = []

        # Séparer l'opérateur de comparaison et la version cible
        operator_regex = re.compile(regex.segment_version)
        if not version.startswith("==") and not version.startswith(">=") and not version.startswith("<=") and not version.startswith("!=") and not version.startswith("^")  and not version.startswith(">") and not version.startswith("<"):
            version = "=="+version

        if " " in version:
            version = version.replace(" ", "")

        match = operator_regex.match(version)
        operator = match.group('operator')
        target_major = match.group('major')
        target_minor = match.group('minor')
        target_patch = match.group('patch')

        target = target_major
        if target_minor!=None:
            target+= "."+target_minor
        if target_patch!=None:
            target+= "."+target_patch
        """
        # Convertir les parties de la version cible en entiers
        if target_major!=None:
            target_major = int(target_major)
        if target_minor!=None:
            target_minor = int(target_minor)
        if target_patch!=None:
            target_patch = int(target_patch)
        """

        # Parcourir toutes les versions disponibles
        for v in releases:
            # Extraire les différentes parties de la version en utilisant l'expression régulière
            if compare_version(v, operator, target):
                compatible_versions.append(v)

        return compatible_versions

@impmagic.loader(
    {'module':'template', 'submodule':['regex']},
    {'module':'re'}
)
def compare_version(v, operator, target):
    version_regex = re.compile(regex.version)

    target_version = version_regex.match(target)
    if target_version!=None:
        #match_release = release_version.groups()

        target_major = target_version.group('major')
        target_minor = target_version.group('minor')
        target_patch = target_version.group('patch')
        # Convertir les parties de la version en entiers
        if target_major!=None:
            target_major = int(target_major)
        if target_minor!=None:
            target_minor = int(target_minor)
        if target_patch!=None:
            target_patch = int(target_patch)
        release_version = version_regex.match(v)
        if release_version!=None:
            #match_release = release_version.groups()
            major = release_version.group('major')
            minor = release_version.group('minor')
            patch = release_version.group('patch')
            # Convertir les parties de la version en entiers
            if major!=None:
                major = int(major)
            if minor!=None:
                minor = int(minor)
            if patch!=None:
                patch = int(patch)

            # Vérifier si la version est compatible en utilisant l'opérateur de comparaison
            if operator == '>=' and (major > target_major or (minor!=None and target_minor!=None and (target_patch==None or patch==None) and major == target_major and minor >= target_minor) or (minor!=None and patch!=None and target_minor!=None and target_patch!=None and major == target_major and minor > target_minor) or (patch!=None and target_patch!=None and major == target_major and minor == target_minor and patch >= target_patch)):
                return True
            elif operator == '>' and (major > target_major or (minor!=None and target_minor!=None and major == target_major and minor > target_minor) or (target_patch!=None and major == target_major and minor == target_minor and patch > target_patch)):
                return True
            elif operator == '<' and (major < target_major or (minor!=None and target_minor!=None and major == target_major and minor < target_minor) or (target_patch!=None and major == target_major and minor == target_minor and patch < target_patch)):
                return True
            elif operator == '<=' and (major < target_major or (minor!=None and target_minor!=None and (target_patch==None or patch==None) and major == target_major and minor <= target_minor) or (minor!=None and patch!=None and target_minor!=None and target_patch!=None and major == target_major and minor < target_minor) or (patch!=None and target_patch!=None and major == target_major and minor == target_minor and patch <= target_patch)):
                return True
            elif operator == '^' and major == target_major:
                return True
            elif operator == '==' and major == target_major and (target_minor is None or minor == target_minor) and (target_patch is None or patch == target_patch):
                return True
            elif operator == '!=' and (major != target_major or minor != target_minor or patch != target_patch):
                return True
    return False


@impmagic.loader(
    {'module':'sys_nxs.connect', 'submodule':['get']},
    {'module':'html5lib.html5parser', 'submodule':['parse']}
)
def search_package(query, strict=None):
        url = f"{__main__.nxs.metadata_source}/"
        
        results = {}

        search = {"q": query}
        response = get(url+"search", params=search)

        content = parse(response.content, namespaceHTMLElements=False)
        for result in content.findall(".//*[@class='package-snippet']"):
            name_element = result.find("h3/*[@class='package-snippet__name']")
            version_element = result.find("h3/*[@class='package-snippet__version']")

            if name_element is None or version_element is None or not name_element.text or not version_element.text:
                continue

            description_element = result.find("p[@class='package-snippet__description']")
            if description_element is not None and description_element.text:
                description = description_element.text
            else:
                description = ""

            try:
                name = name_element.text
                if strict==True and (query not in name and query.replace("_","-") not in name):
                    continue
                results[name] = {}
                results[name]['version'] = version_element.text
                results[name]['description'] = description.strip()
            except:
                pass

        return results

#Vérifie si un package est disponible dans les repos
@impmagic.loader(
    {'module':'sys_nxs.connect', 'submodule':['get']}
)
def in_repo(package_name):
    url = f"{__main__.nxs.metadata_source}/pypi/{package_name}/json"
    response = get(url)
    if response.status_code == 200:
        return True
    elif response.status_code == 404:
        return False

#Récupérer les informations d'un package
@impmagic.loader(
    {'module':'subprocess'},
    {'module':'re'},
    {'module':'sys_nxs.connect', 'submodule':['get']},
    {'module':'app.display', 'submodule':['logs', 'print_nxs']},
    {'module':'structure.check', 'submodule':['in_project', 'get_nexus_file']},
    {'module':'toml_nxs.toml', 'submodule':['TOML'], 'as': 'TOMLnxs'},
    {'module': 'sys_nxs.host', 'submodule': ['path_rep']},
    {'module': 'os.path', 'submodule': ['exists']},
    {'module':'env_nxs.env', 'as':'env'},
    {'module':'template', 'submodule':['regex']}
)
def info_package(package_name, noarbo=False, arbolvl=1, list_version=None):
    package_compiled = re.compile(regex.package_regex)
    package_match = package_compiled.search(package_name)
    if package_match!=None:
        package_name = package_match.group('name')
        version = package_match.group('version')

        if version!=None:
            version = version.replace("==","")
            url = f"{__main__.nxs.metadata_source}/pypi/{package_name}/{version}/json"
        else:
            url = f"{__main__.nxs.metadata_source}/pypi/{package_name}/json"
        
        response = get(url)
        if response.status_code == 200:
            data = response.json()
            info = data['info']

            print_nxs("Name: ", nojump=True)
            print(info['name'])
            if 'version' in info and info['version']!=None and len(info['version'])>0:
                print_nxs("Version: ", nojump=True)
                print(info['version'])
            if 'summary' in info and info['summary']!=None and len(info['summary'])>0:
                print_nxs("Description: ", nojump=True)
                print(info['summary'])
            if 'author' in info and info['author']!=None and len(info['author'])>0:
                print_nxs("Author: ", nojump=True)
                print(info['author'])
            if 'license' in info and info['license']!=None and len(info['license'])>0:
                print_nxs("License: ", nojump=True)
                print(info['license'])
            if 'home_page' in info and info['home_page']!=None and len(info['home_page'])>0:
                print_nxs("Homepage: ", nojump=True)
                print(info['home_page'])

            if not noarbo:
                res = get_requirement(package_name, info['version'])

                if len(res)>0:
                    print_nxs("Dependencies: ")
                    tree_dependencies(res, max=len(res), lvl_max=arbolvl)
                else:
                    print_nxs("Dependencies: ", nojump=True)
                    print("No dependencies")

            #print(res)

            print_nxs("Vulnerabilities: ", nojump=True)
            vuln = data['vulnerabilities']
            if len(vuln)>0:
                print(vuln)
            else:
                print("No vulnerabilities")

            print_nxs("Install: ", nojump=True)
            if in_project():
                tfile = TOMLnxs(get_nexus_file())
                tfile.load_doc()
            else:
                tfile = None

            env_exe = env.get_executable(tfile, logs_env=False)
            if env_exe!=None:
                env_exe = env_exe.replace(path_rep[1], path_rep[0])
                if exists(env_exe):
                    cmd = [env_exe, '-c', 'import pkg_resources; print(pkg_resources.get_distribution("'+package_name+'").location)']
                    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    stdout, stderr = proc.communicate()

                    if len(stdout):
                        cmd = [env_exe, '-c', 'import pkg_resources; print(pkg_resources.get_distribution("'+package_name+'")._version)']
                        proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                        stdout_version, stderr_version = proc.communicate()

                        if len(stdout_version):
                            if stdout_version.rstrip().decode()==info['version']:
                                print("is installed")
                            else:
                                print(f"is installed in another version ({stdout_version.rstrip().decode()})")
                            
                            print_nxs("Location: ", nojump=True)
                            print(stdout.rstrip().decode())
                        else:
                            print("not installed")
                    else:
                        print("not installed")

            if list_version:
                print_nxs("Version available: ")
                for release, rel_info in data['releases'].items():
                    if len(rel_info):
                        print(f"  {release} ({rel_info[0]['upload_time'].replace('T',' ')})")
                    else:
                        print(f"  {release}")

        elif response.status_code == 404:
            logs("package not found", "error")

#Récupérer la dernière version disponible d'un package
@impmagic.loader(
    {'module':'sys_nxs.connect', 'submodule':['get']}
)
def latest_version(package_name):
    url = f"{__main__.nxs.metadata_source}/pypi/{package_name}/json"
    response = get(url)
    if response.status_code == 200:
        data = response.json()
        info = data['info']

        if 'version' in info and len(info['version'])>0:
            return info['version']
    return None

#Vérifie si le package a des vulnérabilités
@impmagic.loader(
    {'module':'sys_nxs.connect', 'submodule':['get']}
)
def is_vulnerable(packages):
    url = f"{__main__.nxs.metadata_source}/pypi/{packages}/json"
    response = get(url)
    if response.status_code == 200:
        data = response.json()
        if len(data['vulnerabilities'])>0:
            return True
    return False

#Affichage de l'arbre de dependance
@impmagic.loader(
    {'module':'app.display', 'submodule':['print_nxs']}
)
def tree_dependencies(packages, lvl=0, lvl_max=1, max=1, arbo=[]):
    i=1
    for package, info in packages.items():
        print_pattern(i, max, arbo)
        if is_vulnerable(package):
            print_nxs(" "+package, color=__main__.color['dark_gray'], nojump=True)
            print_nxs(" (vuln)", color=__main__.color['red'])
        else:
            print_nxs(" "+package, color=__main__.color['dark_gray'])

        if info['version']=="N.A":
            info['version']=">=0.0"
        compatible_version = get_compatible_versions(package, info['version'])

        if compatible_version and len(compatible_version):
            res = get_requirement(package, compatible_version[-1])

            if len(res)>0 and lvl<lvl_max:
                if i==max-1:
                    new_arbo = arbo+['   ']
                else:
                    new_arbo = arbo+['│  ']
                tree_dependencies(res, lvl+1, max=len(res), lvl_max=lvl_max, arbo=new_arbo)
        i+=1

def print_pattern(index, max, arbo):
    arbo = "".join(arbo)
    if index==max-1:
        print(f"{arbo}└─", end="")
    else:
        print(f"{arbo}├─", end="")


#Cherche dans les alias
@impmagic.loader(
    {'module':'json', 'submodule': ['load']},
    {'module':'structure.check', 'submodule': ['get_root_project']},
    {'module':'app.display', 'submodule':['logs']},
    {'module':'os.path', 'submodule': ['exists', 'isfile', 'join', 'abspath']}
)
def get_alias(init=False):
    if init:
        file = abspath('.package_alias')
    else:
        root_project = get_root_project()
        if root_project is False:
            file = '.package_alias'
        else:
            file = join(get_root_project(), '.package_alias')

    if exists(file) and isfile(file):
        try:
            #logs("Chargement des alias de package")
            with open(file) as file:
                return load(file)
        except:
            logs("Fichier corrompu", "warning")
            return {}
    else:
        return {}


#Sauvegarde les alias
@impmagic.loader(
    {'module':'json', 'submodule': ['dump']},
    {'module':'structure.check', 'submodule': ['get_root_project']},
    {'module':'os.path', 'submodule': ['exists', 'isfile', 'join']}
)
def set_alias(data):
    root_project = get_root_project()
    if root_project is False:
        file = '.package_alias'
    else:
        file = join(get_root_project(), '.package_alias')

    with open(file, 'w') as file:
        return dump(data, file, indent=4)

@impmagic.loader(
    {'module':'tomlkit'}
)
def dependency_to_lock(deps, lockfile):
    for package, version in deps.copy().items():
        if isinstance(version, tomlkit.items.Table):
            if package in lockfile:
                deps[package] = dependency_to_lock(version, lockfile[package])
        else:
            if package in lockfile:
                deps[package] = "=="+lockfile[package]   
    return deps


@impmagic.loader(
    {'module':'structure.check', 'submodule': ['get_root_project']}, 
    {'module':'toml_nxs.toml', 'submodule': ['TOML'], 'as': 'TOMLnxs'},
    {'module':'os.path', 'submodule':['join', 'exists']}
)
def lock_package(deps):
    #Si fichier de lock existe
    lock = join(get_root_project(), 'nexus.lock')
    if exists(lock):
        lockfile = TOMLnxs(join(get_root_project(), 'nexus.lock'))
        lockfile.load_doc()
        lock_deps = lockfile.get_key('dependencies')
        if lock_deps!=None:
            deps = dependency_to_lock(deps, lock_deps)

    return deps

#Récupère le dictionnaire des dépendances et enlève les obligations de versions pour tous les packages où uniquement ceux dans name
@impmagic.loader(
    {'module':'tomlkit'}
)
def unversion_package(deps, name=None):
    result = {}
    for element, value in deps.items():
        if isinstance(value, tomlkit.items.Table):
            value = unversion_package(value, name)
            result[element] = value
        else:
            if name==None or (name!=None and name==element):
                result[element] = "N.A"
            else:
                result[element] = value

    return result


#Trouver la version suivante à mettre à jour
@impmagic.loader(
    {'module':'re'},
    {'module':'template', 'submodule':['regex']},
    {'module':'os'}
)
def get_possible_version(deps, actual, minor=False, patch=False):
    #Passage en l'ensemble des noms en lower() pour faciliter la recherche
    actual = [x.lower() for x in actual]

    to_return = {}
    reg_version = re.compile(regex.version)

    for package_name, info in deps.items():
        if package_name=="windows":
            if os.name=="nt":
                win = get_possible_version(info, actual, minor, patch)
                to_return.update(win)
        elif package_name=="linux":
            if os.name!="nt":
                lin = get_possible_version(info, actual, minor, patch)
                to_return.update(lin)
        else:
            if info=="N.A":
                releases = get_versionlist_package(package_name)

                if not minor and not patch:
                    to_return[package_name] = releases[-1]
                else:
                    valid_release = None
                    #match = reg_version.match(actual[package_name.lower()])
                    match = reg_version.match(actual[package_name.lower()])
                    actual_major = match.group('major')
                    actual_minor = match.group('minor')
                    actual_patch = match.group('patch')

                    for release_version in releases:
                        match = reg_version.match(release_version)
                        if match:
                            release_major = match.group('major')
                            release_minor = match.group('minor')
                            release_patch = match.group('patch')

                            if minor:
                                if actual_major==release_major:
                                    valid_release=release_version
                            elif patch:
                                if actual_major==release_major and actual_minor==release_minor:
                                    valid_release=release_version

                    if valid_release==None:
                        valid_release = actual[package_name.lower()]

                    to_return[package_name] = valid_release
            else:
                to_return[package_name] = info

    return to_return


#Retourne uniquement les packages dont la version à changer
def filter_update(actual, deps):
    #Passage en l'ensemble des noms en lower() pour faciliter la recherche
    c_actual = actual.copy()
    actual = {}
    for name, info in c_actual.items():
        actual[name.lower()] = info


    for package_name, package_info in deps.copy().items():
        if package_name.lower() in actual and package_info['version']==actual[package_name.lower()]:
            del deps[package_name]

    return deps


from pony.orm import Required, Optional, PrimaryKey


def Table(db, type_db="sqlite"):
    #Modèle pour le stockage de logs des workers de mise à jour
    class PackageCache(db.Entity):
        id = PrimaryKey(int, auto=True)
        name = Required(str)
        version = Required(str)
        requires_python = Optional(str)
        requires_dist = Optional(str)
        summary = Optional(str)
        author = Optional(str)
        maintainer = Optional(str)
        license = Optional(str)
        home_page = Optional(str)
        vulnerabilities = Optional(str)



import sqlite3
import json
import os
import impmagic

class Cache:
    @impmagic.loader(
        {'module': 'cache.model','submodule': ['Table']},
        {'module': 'pony.orm','submodule': ['Database']},
        {'module': 'os.path','submodule': ['exists', 'isdir', 'dirname']},
        {'module': 'os','submodule': ['makedirs']},
    )
    def __init__(self, nom_base_de_donnees):
        self.orm = Database()
        Table(self.orm)

        if not exists(dirname(nom_base_de_donnees)) or not isdir(dirname(nom_base_de_donnees)):
            makedirs(dirname(nom_base_de_donnees))

        self.orm.bind(provider='sqlite', filename=nom_base_de_donnees, create_db=True)
        self.orm.generate_mapping(create_tables=True)


    @impmagic.loader(
        {'module':'app.display', 'submodule': ['logs']},
        {'module': 'pony.orm','submodule': ['db_session']},
    )
    def add(self, element):
        try:
            with db_session:
                self.orm.PackageCache(**element)

                return True
        except Exception as err:
            logs(err, "critical")
            return False


    @impmagic.loader(
        {'module': 'pony.orm','submodule': ['db_session']},
    )
    def remove_by_id(self, element_id):
        try:
            with db_session:
                element = self.orm.PackageCache.get(id=element_id)
                if element:
                    element.delete()
                    return True
                return False

        except Exception as err:
            logs(err, "critical")
            return False


    @impmagic.loader(
        {'module': 'pony.orm','submodule': ['db_session', 'select']},
        {'module':'app.display', 'submodule': ['logs']},
        {'module':'package.package', 'submodule': ['compare_version']}, 
    )
    def get_package_from_cache(self, package_name, version, operator):
        try:
            query = select(element for element in self.orm.PackageCache)

            query = query.filter(lambda element: element.name == package_name.lower().replace("-", "_"))

            if version and operator:
                results = []
                for element in query:
                    if compare_version(element.version, operator, version):
                        results.append(element)
                return results
            else:
                return query[:]

        except Exception as err:
            logs(err, "critical")
            return []

    @impmagic.loader(
        {'module': 'pony.orm','submodule': ['db_session', 'select']},
        {'module':'app.display', 'submodule': ['logs']},
    )
    def remove_package(self, package_name, version=None, operator="=="):
        try:
            with db_session:
                extract = self.get_package_from_cache(package_name, version, operator)

                if extract:
                    for line in extract:
                        line.delete()
                        logs(f"Package {package_name}=={line.version} supprimé")

                else:
                    logs("Aucun élément trouvé")

        except Exception as err:
            logs(err, "critical")
            return False


    @impmagic.loader(
        {'module': 'json'},
        {'module': 'pony.orm','submodule': ['db_session', 'select']},
        {'module':'app.display', 'submodule': ['logs']},
    )
    def search(self, package_name, version=None, operator="=="):
        result = []

        with db_session:
            extract = self.get_package_from_cache(package_name, version, operator)

        if extract:
            for line in extract:
                for index in ['name', 'version', 'requires_python', 'requires_dist', 'summary', 'author', 'maintainer', 'license', 'home_page', 'vulnerabilities']:
                    try:
                        converted = json.loads(getattr(line, index))
                        setattr(line, index, converted)
                    except Exception as err:
                        pass

                insert = {'name': line.name,'version': line.version,'requires_python': line.requires_python,'requires_dist': line.requires_dist,'summary': line.summary,'author': line.author,'maintainer': line.maintainer,'license': line.license,'home_page': line.home_page, 'vulnerabilities': line.vulnerabilities}
                result.append(insert)

        
        return result


    @impmagic.loader(
        {'module': 'pony.orm','submodule': ['db_session', 'select']},
        {'module':'app.display', 'submodule': ['logs']},
    )
    def select_all(self):
        try:
            with db_session:
                #log = __main__.orm_deploy.Argos_Server.get(name_server=self.hostname)
                query = select(log for log in self.orm.PackageCache)[:]
                return query
        except Exception as err:
            logs(err, "critical")
            return []



classic = {'cyan': 'cyan', 'green': 'green', 'dark_gray': 'dark_gray', 'light_gray': 'light_gray', 'yellow': 'yellow', 'red': 'red', 'light_red': 'light_red', 'orange_1': 'orange_1'}
truecolor = {'cyan': '26,128,162', 'green': 'green', 'dark_gray': 'dark_gray', 'light_gray': 'light_gray', 'yellow': 'light_goldenrod_2b', 'red': 'red', 'light_red': 'light_red', 'orange_1': 'red'}



import sys
from template.regex import project_version, valid_file, python_version, dependencies

DEFAULT_TOML = {
	"projectname": {"name": "name", "section": "project", "type": str, "default_value": ""},
	"version": {"name": "version", "section": "project", "type": str, "regex": project_version,"default_value": "0.1.0"},
	"description": {"name": "description", "section": "project", "type": str, "default_value": ""},
	"mainfile": {"name": "mainfile", "section": "project", "type": str, "regex": valid_file,"default_value": ""},
	"pythonversion": {"name": "requires-python", "section": "project", "type": str, "regex": python_version,"default_value": f">={sys.version_info.major}.{sys.version_info.minor}"},
	"authors": {"name": "authors", "section": "project.metadata", "type": list,"default_value": []},
	"maintainers": {"name": "maintainers", "section": "project.metadata", "type": list, "default_value": []},
	"readme": {"name": "readme", "section": "project.metadata", "type": str, "default_value": "README.md"},
	"changelog": {"name": "changelog", "section": "project.metadata", "type": str, "default_value": "changelog.md"},
	"license": {"name": "license", "section": "project.metadata", "type": str, "default_value": ""},
	"license_file": {"name": "license_file", "section": "project.metadata", "type": str, "default_value": ""},
	"keywords": {"name": "keywords", "section": "project.metadata", "type": list, "default_value": []},
	"classifiers": {"name": "classifiers", "section": "project.metadata", "type": list, "default_value": []},
	"copyright": {"name": "copyright", "section": "project.metadata", "type": str, "default_value": ""},
	"dependencies": {"name": "dependencies", "section": "project", "type": dict, "regex": dependencies,"default_value": {}},
	"dependencies_win": {"name": "windows", "section": "project.dependencies", "type": dict, "regex": dependencies, "default_value": {}},
	"dependencies_linux": {"name": "linux", "section": "project.dependencies", "type": dict, "regex": dependencies, "default_value": {}},
	"package": {"name": "packages", "section": "project.build", "type": list, "default_value": []},
	"include": {"name": "includes", "section": "project.build", "type": list, "default_value": []},
	"exclude": {"name": "excludes", "section": "project.build", "type": list, "default_value": []},
	"include_files": {"name": "include_files", "section": "project.build", "type": list, "default_value": []},
	"optimize": {"name": "optimize", "section": "project.build", "type": int, "default_value": 0},
	"compress": {"name": "no_compress", "section": "project.build", "type": bool, "default_value": False},
	"gui": {"name": "GUI", "section": "project.build", "type": bool, "default_value": False},
	"icon": {"name": "icon", "section": "project.build", "type": str, "default_value": ""},
	"platforms": {"name": "platforms", "section": "project.wheel", "type": str, "default_value": "ALL"},
	"console_scripts": {"name": "console_scripts", "section": "project.wheel.entry_points", "type": list, "default_value": []},
	"homepage": {"name": "homepage", "section": "project.urls", "type": str, "default_value": ""},
	"documentation": {"name": "documentation", "section": "project.urls", "type": str, "default_value": ""},
	"changelog": {"name": "changelog", "section": "project.urls", "type": str, "default_value": ""}
}


import __main__
import impmagic

@impmagic.loader(
    {'module': 'zpp_color', 'submodule': ['fg', 'attr']}
)
def prompt(name, default=None):
	default_display = default
	if (isinstance(default_display, list) or isinstance(default_display, dict)) and len(default_display)==0:
		default_display = ""
	if default!=None:
		inp = input(f"{fg(__main__.color['cyan'])}{name} [{fg('green')}{default_display}{fg(__main__.color['cyan'])}]: {attr(0)}")
		if inp=="":
			return default
	else:
		inp = input(f"{fg(__main__.color['cyan'])}{name}: {attr(0)}")
	return inp


@impmagic.loader(
    {'module': 'pynput.keyboard', 'submodule': ['Controller']}
)
def form(name, default=None):
	print(f"{default}: ", end="")
	if default:
		keyboard = Controller()
		keyboard.type(default)
	return input()


@impmagic.loader(
    {'module': 'zpp_color', 'submodule': ['fg', 'attr']},
    {'module': 'datetime', 'submodule': ['datetime']},
    {'module': 'template', 'submodule': ['default_conf']}
)
def logs(message, lvl='info', nodate=None):
	if __main__.nxs.conf.load(val='logs.display', section='',default=default_conf.logs_display):
		if lvl=='logs':
			color = __main__.color['light_gray']
		elif lvl=='info':
			color = __main__.color['cyan']
		elif lvl=='warning':
			color = __main__.color['yellow']
		elif lvl=='error':
			color = __main__.color['red']
		elif lvl=='critical':
			color = __main__.color['light_red']
		elif lvl=='valid' or lvl=='success':
			color = __main__.color['green']
		
		if nodate==False or (nodate==None and __main__.nxs.conf.load(val='logs.date', section='',default=default_conf.logs_date)):
			date = datetime.now().strftime("%Y/%m/%d - %H:%M:%S.%f")
			print(f"{fg(__main__.color['dark_gray'])}[{date}] - {attr(0)}{fg(color)}{message}{attr(0)}")
		else:
			print(f"{fg(color)}{message}{attr(0)}")

@impmagic.loader(
    {'module': 'zpp_color', 'submodule': ['fg', 'attr']}
)
def print_nxs(message, color=None, nojump=False):
	if color==None:
		color = __main__.color['cyan']
	
	if nojump:
		print(f"{fg(color)}{message}{attr(0)}", end="")
	else:
		print(f"{fg(color)}{message}{attr(0)}")


@impmagic.loader(
    {'module': 'tempfile','submodule': ['NamedTemporaryFile']},
    {'module': 'json','submodule': ['dumps']},
    {'module': 'pygments.lexers','submodule': None, 'as':'lexers'},
    {'module': 'pygments.formatters','submodule': ['TerminalTrueColorFormatter']},
    {'module': 'pygments','submodule': ['highlight']}
)
def cat_setup(content):
    BUF_SIZE = 65536

    f = NamedTemporaryFile()
    f.write(dumps(content, indent=4).encode())
    f.seek(0)

    try:
        lex = lexers.get_lexer_by_name('json')
        formatter = TerminalTrueColorFormatter(style="monokai")
        
        while True:
            line = f.readline() 
            if not line:
                break
            else:
                print(highlight(line, lex, formatter),end="")
        f.close()
    except:
        with open(file, 'rb') as f:
            while True:
                data = f.read(BUF_SIZE).decode('utf-8', 'ignore').translate({ord('\u0000'): None})
                if not data:
                    break
                print(data)

def bytes2human(n):
	symbols = ('K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
	prefix = {}
	for i, s in enumerate(symbols):
		prefix[s] = 1 << (i + 1) * 10
	for s in reversed(symbols):
		if n >= prefix[s]:
			value = float(n) / prefix[s]
			return '%.1f%s' % (value, s)
	return "%sB" % n


import impmagic

@impmagic.loader(
    {'module':'black', 'submodule': ['FileMode', 'format_file_contents']},
    {'module':'black.report', 'submodule': ['NothingChanged']},
    {'module':'difflib', 'submodule': ['unified_diff']},
    {'module':'os.path', 'submodule': ['exists']},
    {'module':'Levenshtein'}
)
def analyse(filename, apply=False, fileout=None, fast=False):
    if exists(filename):
        with open(filename) as f:
            content = f.read()

            try:
                mode = FileMode()
                out = format_file_contents(content, fast=fast, mode=mode)
                
                result = ""
                for line in unified_diff(content.splitlines(), out.splitlines(), lineterm='', fromfile=filename, tofile='ClearCode', n=1):
                    if line.startswith("@@") and line.endswith("@@"):
                        result+="\n\n"+line+"\n"
                    else:
                        result+=line+"\n"
                
                cat_content(result)
                
                distance = Levenshtein.distance(content, out)
                similarity = 1 - (distance / max(len(content), len(out)))
                print("\nTaux de similarité: {:.2f}%\n".format(similarity * 100))

            except NothingChanged:
                print(f"Fichier {filename} déjà optimisé\n")

        if apply:
            if fileout==None:
                fileout = filename

            with open(fileout, 'w') as f:
                f.write(out)

    else:
        print(f"Fichier {filename} introuvable\n")



@impmagic.loader(
    {'module': 'pygments.lexers','submodule': None, 'as':'lexers'},
    {'module': 'pygments.formatters','submodule': ['TerminalTrueColorFormatter']},
    {'module': 'pygments','submodule': ['highlight']}
)
def cat_content(file):
    BUF_SIZE = 65536

    try:
        lex = lexers.guess_lexer(file)
        formatter = TerminalTrueColorFormatter(style='monokai')
        
        print(highlight(file, lex, formatter),end="")
    except:
        data = file.decode('utf-8', 'ignore').translate({ord('\u0000'): None})
        print(data)


import __main__
import impmagic

class DirectCache:
	def __init__(self):
		if not hasattr(__main__.nxs, 'cache'):
			__main__.nxs.cache = {}

	def get(self, type_cache, name):
		name = name.lower().replace("-","_")

		if type_cache not in __main__.nxs.cache.keys():
			return None
		else:
			if name in __main__.nxs.cache[type_cache]:
				return __main__.nxs.cache[type_cache][name]
			else:
				return None

	def set(self, type_cache, name, content):
		name = name.lower().replace("-","_")

		if type_cache not in __main__.nxs.cache.keys():
			__main__.nxs.cache[type_cache] = {}

		__main__.nxs.cache[type_cache][name] = content


import impmagic
from os import name

#Calcul du chemin d'un fichier
@impmagic.loader(
    {'module':'os.path', 'submodule':['isabs','abspath']}
)
def path_reg(arg):
	if isabs(arg):
		return arg
	return abspath(arg)

#Cherche le terminal (cmd,powershell,bash...) sur lequel est exécuté l'app
@impmagic.loader(
    {'module':'os'},
    {'module':'psutil'}
)
def get_os():
	if os.name=="nt":
		if "nxs" not in psutil.Process(os.getppid()).name():
			return psutil.Process(os.getppid()).name()
		else:
			return psutil.Process(psutil.Process(os.getppid()).ppid()).name()
	else:
		return os.popen(f'ps -p {os.getppid()} -o comm=').read().strip()

if name=='nt':
    path_rep = ["\\","/"]
else:
    path_rep = ["/","\\"]


import impmagic

def add_key(dictionary, key_string, value):
    # Séparation des différents éléments du string
    elements = key_string.split(".")
    # Ajout de la clé au dictionnaire à l'aide de la méthode update
    current_dict = dictionary
    for element in elements[:-1]:
        current_dict = current_dict.setdefault(element, {})
    current_dict.update({elements[-1]: value})
    return dictionary


class TOML:
    def __init__(self, file):
        self.file = file

    @impmagic.loader(
        {'module':'tomlkit'},
        {'module':'os.path', 'submodule':['exists']}
    )
    def load_doc(self):
        if exists(self.file):
            with open(self.file, 'r') as f:
                self.doc = tomlkit.parse(f.read())
        else:
            self.doc = tomlkit.document()

    def get_key(self, key, section=None):
        if hasattr(self, "doc"):
            if section is None:
                #Cas où section est None (éléments sans section)
                return self.doc.get(key, None)

            #return self.doc[section].get(key, None)

            section_levels = section.split('.')
    
            # Parcours de chaque niveau de section
            current_section = self.doc
            for level in section_levels:
                # Récupération de la sous-section correspondante
                current_section = current_section.get(level, None)
                if current_section is None:
                    # Si la sous-section n'existe pas, on retourne None
                    return None
            
            # Récupération de la valeur de la clé dans la section cible
            return current_section.get(key, None)

        return None

    @impmagic.loader(
        {'module':'tomlkit'}
    )
    def dict_to_toml(self, data):
        self.doc = tomlkit.document()

        for key, value in data.items():
            if isinstance(value, dict):
                table = tomlkit.table()
                for subkey, subvalue in value.items():
                    if isinstance(subvalue, dict):
                        subtable = tomlkit.table()
                        for subsubkey, subsubvalue in subvalue.items():
                            subtable.add(subsubkey, subsubvalue)
                        table.add(subkey, subtable)
                    else:
                        table.add(subkey, subvalue)
                self.doc.add(key, table)
            else:
                self.doc.add(key, value)

        self.write(self.doc)

    @impmagic.loader(
        {'module':'template.toml_format', 'submodule':['DEFAULT_TOML']},
        {'module':'structure.license', 'submodule':['licenses_availables']},
        {'module':'re', 'submodule':['match']}
    )
    def new(self, data_user):
        data = {}
        for block in DEFAULT_TOML:
            value = None
            if block in data_user:
                #Condition spécial pour les licenses
                if block=="license":
                    if licenses_availables(data_user[block]):
                        value = data_user[block]
                else:
                    if isinstance(data_user[block], DEFAULT_TOML[block]["type"]):
                        if DEFAULT_TOML[block]["type"]==str and "regex" in DEFAULT_TOML[block]:
                            if match(DEFAULT_TOML[block]["regex"], data_user[block]):
                                value = data_user[block]
                        else:
                            value = data_user[block]

            if value==None:
                value = DEFAULT_TOML[block]["default_value"]

            data = add_key(data, DEFAULT_TOML[block]["section"]+"."+DEFAULT_TOML[block]["name"], value)

        #data = ast.literal_eval(TOML_TEMPLATE.format(**data))

        self.dict_to_toml(data)

    def edit_key(self, key, value, section):
        if hasattr(self, 'doc'):
            self.load_doc()

        sections = section.split('.')  
        current_section = self.doc
        for s in sections:
            if s not in current_section:
                current_section.add(s, {})
            current_section = current_section[s]
        if key not in current_section:
            current_section.add(key, 'default value')

        """
        # Vérification de l'existence de la section et de la clé
        if section not in self.doc:
            self.doc.add(section, {})
        if key not in self.doc[section]:
            self.doc[section].add(key, 'default value')

        # Modification de la valeur de la clé
        self.doc[section][key] = value
        """
        current_section[key] = value

        self.write(self.doc)

    def delete_key(self, key, section):
        if hasattr(self, 'doc'):
            self.load_doc()

        sections = section.split('.')
        current_section = self.doc

        # Parcours des niveaux de section
        for s in sections:
            # Vérification de l'existence de la section
            if s not in current_section:
                return False  # La section n'existe pas, retourne False
            current_section = current_section[s]

        # Vérification de l'existence de la clé
        if key not in current_section:
            return False  # La clé n'existe pas, retourne False

        # Suppression de la clé
        del current_section[key]

        self.write(self.doc)  # Écriture des modifications dans le fichier
        return True


    def write(self, data):
        data = data.as_string()

        while "\n\n\n" in data:
            data = data.replace("\n\n\n","\n")

        if "]\n[" in data:
            data = data.replace("]\n[","]\n\n[")

        with open(self.file, 'w') as f:
            f.write(data)


def read_data(data, pattern=""):
    if isinstance(data, dict):
        for key, value in data.items():
            if isinstance(value, list) or isinstance(value, dict):
                print(f"{pattern}{key}:")
                read_data(value, pattern+"  ")
            elif isinstance(value, str):
                print(f"{pattern}{key}: {value}")
    elif isinstance(data, list):
        print(f"{pattern}{list(data)}")
        """
        for element in data:
            if isinstance(element, list) or isinstance(element, dict):
                read_data(element, pattern+"  ")
            elif isinstance(element, str):
                print(f"{pattern}{element}")
        """
    elif isinstance(data, str):
        print(f"{pattern}{data}")
    elif isinstance(data, int) or isinstance(data, float):
        print(f"{pattern}{data}")



import impmagic

#Controle que la clé sont présente et qu'elles respectent les regex
@impmagic.loader(
    {'module':'toml_nxs.toml', 'submodule':['TOML'], 'as':'TOMLnxs'},
    {'module':'template.toml_format', 'submodule':['DEFAULT_TOML']},
    {'module':'re', 'submodule':['match']},
    {'module':'app.display', 'submodule':['logs']}
)
def is_nexus_project(command, nolog=False):
	if in_project():
		parameter_mandatory = []
		if command=='run':
			parameter_mandatory = ['projectname', 'mainfile']
		elif command=='version':
			parameter_mandatory = ['version']
		elif command=='project':
			parameter_mandatory = ['project']
		elif command=='lock':
			parameter_mandatory = ['dependencies']
		elif command=='env':
			parameter_mandatory = ['projectname']
		elif command=='add' or command=='remove':
			parameter_mandatory = ['dependencies', 'package', 'include', 'include_files']
		elif command=='check':
			parameter_mandatory = ['dependencies', 'mainfile', 'package', 'include']
		elif command=='backup':
			parameter_mandatory = ['projectname','include_files','license_file','readme']
		elif command=='pack' or command=='compile':
			parameter_mandatory = ['version','mainfile','projectname']
		elif command=='pack':
			parameter_mandatory = ['projectname','version','description','authors','maintainers','keywords', 'classifiers','platforms','homepage','documentation','changelog']

		tfile = TOMLnxs(get_nexus_file())
		tfile.load_doc()

		for param in parameter_mandatory:
			res = tfile.get_key(DEFAULT_TOML[param]['name'], DEFAULT_TOML[param]['section'])
			if res==None:
				logs(f"Clé {DEFAULT_TOML[param]['name']} non trouvée", "warning")
				return False
			else:
				if "regex" in DEFAULT_TOML[param] and not isinstance(res, dict):
					if not match(DEFAULT_TOML[param]["regex"], res):
						logs(f"Clé {DEFAULT_TOML[param]['name']} non valide", "warning")
						return False
		return True
	else:
		if not nolog:
			logs("Vous n'êtes pas dans un projet","critical")
	return False

@impmagic.loader(
    {'module':'os.path', 'submodule':['abspath', 'exists', 'join', 'dirname']},
    {'module':'app.display', 'submodule':['logs']},
    {'module':'sys', 'submodule':['exit']},
    {'module':'template', 'submodule':['default_conf']},
    {'module':'__main__'} 
)
def get_nexus_file(log=True, _break=True):
	deep = __main__.nxs.conf.load(val='project.deep_search', section='',default=default_conf.project_deep_search)

	folder = abspath('.')

	i=0
	while i!=deep:
		file = join(folder,'nexus.toml')
		if exists(file):
			return file
		else:
			folder = dirname(folder)
		i+=1
	
	if log:
		logs("Aucun fichier nexus.toml trouvé", "critical")

	if _break:
		exit()


@impmagic.loader(
    {'module':'os.path', 'submodule':['abspath', 'exists', 'join', 'dirname']},
    {'module':'template', 'submodule':['default_conf']},
    {'module':'__main__'}
)
def in_project(deep = None):
	if deep==None:
		deep = __main__.nxs.conf.load(val='project.deep_search', section='',default=default_conf.project_deep_search)

	folder = abspath('.')

	i=0
	while i!=deep:
		file = join(folder,'nexus.toml')
		if exists(file):
			return True
		else:
			folder = dirname(folder)
		i+=1
	
	return False


@impmagic.loader(
    {'module':'os.path', 'submodule':['abspath', 'exists', 'join', 'dirname']},
    {'module':'template', 'submodule':['default_conf']},
    {'module':'__main__'}
)
def get_root_project(deep = None):
	if deep==None:
		deep = __main__.nxs.conf.load(val='project.deep_search', section='',default=default_conf.project_deep_search)

	folder = abspath('.')

	i=0
	while i!=deep:
		file = join(folder,'nexus.toml')
		if exists(file):
			return folder
		else:
			folder = dirname(folder)
		i+=1
	
	return False


import impmagic

@impmagic.loader(
    {'module': 'glob', 'submodule': ['glob']},
    {'module': 'os.path', 'submodule': ['exists', 'join', 'isfile', 'isdir']}
)
def construct_pack_data(element, pack_data, codefolder, deep=False):
      file = join(codefolder, element)
      if deep==False and exists(file+".py") and isfile(file+".py"):
            pack_data.append(file.replace(codefolder+path_rep[0],'')+".py")
      elif deep==False and exists(file+".pyw") and isfile(file+".pyw"):
            pack_data.append(file.replace(codefolder+path_rep[0],'')+".pyw")
      elif exists(join(codefolder, element)):
            pack_data.append(element+path_rep[0]+"*")
            for content in glob(join(codefolder, element)+path_rep[0]+"*"):
                  if not content.endswith('__pycache__') and isdir(content):
                        pack_data = construct_pack_data(content.replace(codefolder+path_rep[0],''), pack_data, codefolder, deep=True)
      return pack_data


def construct_platform(dictionary):
      list_package = []
      for package, version in dictionary.items():
            if version!="N.A":
                  list_package.append(package+version)
            else:
                  list_package.append(package)

      return list_package

@impmagic.loader(
    {'module': 'setuptools'},
    {'module': 'os', 'submodule': ['chdir']},
    {'module': 'os', 'submodule': ['name'], 'as':'system_name'},
    {'module': 'os.path', 'submodule': ['exists', 'join', 'isfile']},
    {'module': 'sys'},
    {'module': 'sys_nxs.host', 'submodule': ['path_rep']},
    {'module': 'template.regex', 'submodule': ['author_regex']},
    {'module': 're', 'submodule': ['match']}
)
def build(tfile, projectfolder, version=None, typefile="bdist_wheel", dry_run=False):
      projectname = tfile.get_key('name','project')
      ## README FILE
      long_desc = tfile.get_key('readme','project.metadata')
      if long_desc!="" and exists(long_desc):
            long_desc = open(join(projectfolder, long_desc)).read()
            description_type = 'text/markdown'
      else:
            long_desc = None
            description_type = None

      ## URL INFORMATION
      url = tfile.get_key('homepage','project.urls')
      if url=="":
            url = None

      project_urls = {}
      doc = tfile.get_key('documentation','project.urls')
      if doc!="":
            project_urls['Documentation'] = doc

      changelog = tfile.get_key('changelog','project.urls')
      if changelog!="":
            project_urls['Changelog'] = changelog

      if len(project_urls)==0:
            project_urls={}

      #OTHER OPTIONS
      console_script = tfile.get_key('console_scripts','project.wheel.entry_points')
      if len(console_script)==0:
            entry_points = None
      else:
            entry_points = {
              'console_scripts': console_script,
            }

      platform = tfile.get_key('platforms','project.wheel')
      if platform=="":
            platform = "ALL"

      license = tfile.get_key('license','project.metadata')
      if license=="":
            license=None


      author = tfile.get_key('authors','project.metadata')
      if len(author)>0:
            new = ""
            for element in author:
                  if match(author_regex, element):
                        if new=="":
                              new = element
                        else:
                              new += ", "+element
            author = new
      else: 
            author=None

      maintainer = tfile.get_key('maintainers','project.metadata')
      if len(maintainer)>0:
            new = ""
            for element in maintainer:
                  if match(author_regex, element):
                        if new=="":
                              new = element
                        else:
                              new += ", "+element
            maintainer = new
      else: 
            maintainer=None
      
      includes = tfile.get_key("includes", "project.build")
      include_files = tfile.get_key("include_files", "project.build")
      includable = includes+include_files
      codefolder = join(projectfolder,tfile.get_key('name','project'))
      pack_data = []
      for element in includable:
            pack_data = construct_pack_data(element, pack_data, codefolder)

      if len(pack_data)==0:
            package_data = {}
      else:
            package_data = {}
            package_data[projectname] = pack_data

      if version=="windows":
            platform="win_amd64"
      elif version=="linux":
            platform="manylinux2010_x86_64"

      install_requires = tfile.get_key('dependencies','project').copy() if tfile.get_key('dependencies','project') is not None else None

      if install_requires==None or len(install_requires)==0:
            install_requires=None
      else:
            if version=='windows' and 'windows' in install_requires:
                  requires_platform = install_requires.copy()
                  for element, element_version in install_requires['windows'].items():
                        if element not in install_requires:
                              requires_platform[element] = element_version
                  install_requires = requires_platform

            elif version=='linux' and 'linux' in install_requires:
                  requires_platform = install_requires.copy()
                  for element, element_version in install_requires['linux'].items():
                        if element not in install_requires:
                              requires_platform[element] = element_version
                  install_requires = requires_platform

            if "windows" in install_requires:
                  install_requires.remove('windows')
            if "linux" in install_requires:
                  install_requires.remove('linux')


      chdir(projectfolder)
      #exit()
      #setup.py sdist bdist_wheel --dist-dir=dist
      argument = ['pack.py', typefile, "--dist-dir=dist"]
      if platform!="ALL":
            argument.append("--plat-name="+platform)

      sys.argv = argument

      if install_requires!=None:
            install_requires = construct_platform(install_requires)


      if dry_run:
            from app.display import cat_setup

            if long_desc!=None:
                  long_desc = "-- CONTENT FILE --"

            setup = {'name':projectname,
                  'version': tfile.get_key('version','project'),
                  'author': author,
                  'maintainer': maintainer,
                  'keywords': " ".join(tfile.get_key('keywords','project.metadata')),
                  'classifiers': tfile.get_key('classifiers','project.metadata'),
                  'packages': [tfile.get_key('name','project')],
                  'description': tfile.get_key('description','project'),
                  'long_description': long_desc,
                  'long_description_content_type': description_type,
                  'url': url,
                  'project_urls': project_urls,
                  'entry_points': entry_points,
                  'install_requires': install_requires,
                  'package_data': package_data,
                  'platforms':platform,
                  'license':license}

            cat_setup(setup)

      else:
            setuptools.setup(name=projectname,
                  version=tfile.get_key('version','project'),
                  author=author,
                  maintainer=maintainer,
                  keywords = " ".join(tfile.get_key('keywords','project.metadata')),
                  classifiers = tfile.get_key('classifiers','project.metadata'),
                  packages=[tfile.get_key('name','project')],
                  description=tfile.get_key('description','project'),
                  long_description = long_desc,
                  long_description_content_type=description_type,
                  url = url,
                  project_urls=project_urls,
                  entry_points = entry_points,
                  install_requires = install_requires,
                  package_data=package_data,
                  platforms = platform,
                  license=license)



@impmagic.loader(
    {'module': 'setuptools', 'submodule': ['setup']},
    {'module': 'sys'}
)
def build_fake(projectname, foldername):
      #setup.py sdist bdist_wheel --dist-dir=dist
      sys.argv = ["pack.py", "sdist", "--dist-dir="+foldername]
      setup(name=projectname,
            version="0.0.0")


from contextlib import contextmanager
import impmagic


@impmagic.loader(
    {'module':'sys'}
)
@contextmanager
def ignore_site_packages_paths():
    paths = sys.path
    sys.path = list(filter(
        None,
        filter(lambda i: 'site-packages' not in i, sys.path)
    ))
    yield
    sys.path = paths

@impmagic.loader(
    {'module':'sys'},
    {'module':'inspect'}
)
@contextmanager
def ignore_module(module_name):
    original_module = sys.modules.copy()  # Récupérer le module original ou None
    for element in original_module:
        if element==module_name or element==module_name+".":
            del sys.modules[element]
    try:
        yield
    finally:
        if original_module:
            sys.modules = original_module 

@impmagic.loader(
    {'module':'sys'},
    {'module':'importlib', 'submodule': ['import_module']},
    {'module':'os.path', 'submodule': ['exists', 'isfile']}
)
def is_native_module(module):
    if module in ['pip', 'setuptools']:
        return True
    if exists(module+".py") or exists(module+".pyw"):
        return False
    else:
        if module in sys.builtin_module_names:
            return True
        with ignore_module(module):
            with ignore_site_packages_paths():
                try:
                    import_module(module)
                except ModuleNotFoundError:
                    return False
                except ImportError:
                    return False
                else:
                    return True


import __main__
import impmagic


#Création du context
class Context:
    @impmagic.loader(
        {'module':'sys_nxs.host', 'submodule': ['path_rep']},
        {'module':'os.path', 'submodule': ['dirname']},
        {'module': 'sys'}
    )
    def __init__(self, virtdir, virtname = None, prompt=None):
        self.env_dir = virtdir
        if virtname==None:
            name = virtdir.split(path_rep[0])
            self.env_name = name[len(name)-1]
        else:
            self.env_name = virtname

        if prompt!=None:
            self.prompt = f'({prompt}) '
        else:
            self.prompt = f'({self.env_name}) '

        self.executable = sys.executable
        self.python_dir = dirname(sys.executable)
        self.python_exe = sys.executable.replace(self.python_dir,"").replace(path_rep[0],"")
        if sys.platform == 'win32':
            self.bin_name = 'Scripts'
            self.lib_path = virtdir + path_rep[0] +'Lib'
            self.inc_path = virtdir + path_rep[0] +'Include'
        else:
            self.bin_name = 'bin'
            self.lib_path = virtdir + path_rep[0] +'lib'
            self.inc_path = virtdir + path_rep[0] +'include'
        self.bin_path = virtdir + path_rep[0] +self.bin_name
        self.env_exe = self.bin_path + path_rep[0] + self.python_exe
        self.env_exec_cmd = self.env_exe
        self.cfg_path = virtdir + path_rep[0] + 'pyvenv.cfg'

@impmagic.loader(
    {'module':'os.path', 'submodule': ['exists','isfile','isdir']}
)
def is_valid_envdir(envdir):
    context = Context(envdir)

    for directory in [context.lib_path,context.bin_path]:
        if not exists(directory) or not isdir(directory):
            return None

    for file in [context.env_exe,context.cfg_path]:
        if not exists(file) or not isfile(file):
            return None

    return context
    


#Récupérer le chemin de l'exécutable (root ou de l'environnement default)
@impmagic.loader(
    {'module':'sys'},
    {'module':'os.path', 'submodule': ['join', 'exists', 'isfile', 'isabs']},
    {'module':'app.display', 'submodule': ['logs']},
    {'module': 'sys_nxs.host', 'submodule': ['path_rep']},
    {'module':'structure.check', 'submodule': ['get_root_project']}
)
def get_executable(tfile=None, env_name=None, root=True, logs_env=True):
    if tfile==None:
        return sys.executable

    env_exe=None
    if 'venv' in tfile.doc:
        for env in tfile.doc['venv']:
            name = env
            env = tfile.doc['venv'][env]
            #Prend le default si aucun env est spécifié
            if env_name==None and 'default' in env and env['default']==True:
                env_exe = env['env_exe']
                break
            elif env_name!=None and name==env_name:
                env_exe = env['env_exe']
                break

    if env_exe!=None:
        if logs_env:
            logs("Run in env ")

        if isabs(env_exe):
            executable = env_exe.replace(path_rep[1], path_rep[0])
        else:
            executable = join(get_root_project(), env_exe).replace(path_rep[1], path_rep[0])

        if exists(executable) and isfile(executable):
            return executable
        else:
            if logs_env and env_name==None:
                logs(f"Environnement {name} introuvable", "error")
            elif logs_env and env_name!=None:
                logs(f"Environnement {env_name} introuvable", "error")
            return
    elif env_name!=None and env_exe==None:
        if logs_env:
            logs(f"Environnement {env_name} introuvable", "error")
        return
    elif env_exe==None and root==True:
        if logs_env:
            logs("Run in main python")
        return sys.executable
    else:
        return

#Récupérer les informations de l'environnement (root ou de l'environnement default)
def get_env(tfile, env_name=None, root=True, logs_env=True):
    env_exe=None
    if 'venv' in tfile.doc:
        for env in tfile.doc['venv']:
            name = env
            env = tfile.doc['venv'][env]
            #Prend le default si aucun env est spécifié
            if env_name==None and 'default' in env and env['default']==True:
                env_exe = env
                break
            elif env_name!=None and name==env_name:
                env_exe = env
                break

    if env_exe!=None:
        return env
    elif env_exe==None and root==True:
        return None
    else:
        return None


#Check si le package est déjà installé
@impmagic.loader(
    {'module':'subprocess'},
    {'module':'re', 'submodule': ['compile']},
    {'module':'sys_nxs.host', 'submodule': ['path_rep']},
    {'module':'template.regex', 'submodule': ['dependencies']}
)
def get_package(env_exe, namemodule):
    env_exe = env_exe.replace(path_rep[1], path_rep[0])

    reg = compile(dependencies)
    module = reg.match(namemodule)
    namemodule = module.group(1)
    namemodule = namemodule.lower().replace("-","_")
    #namemodule = namemodule.replace("-","_")

    #Vérifie si le retour commande pip freeze existe déjà (pour éviter de l'appeler plusieurs fois)
    if not hasattr(__main__.nxs, 'already_installed'):
        #cmd = [env_exe, '-c', 'from pip._internal.operations import freeze;print("\\n".join([ package for package in freeze.freeze()]));']
        #cmd = [env_exe, '-c', 'from pkg_resources import working_set;print("\\n".join([package.project_name + "==" + package.version for package in working_set]))']
        cmd = [env_exe, '-m', 'pip', 'freeze']
        proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()
        __main__.nxs.already_installed = stdout
    else:
        stdout = __main__.nxs.already_installed

    for pack in stdout.decode().split("\r\n"):
        if pack!="":
            pack = pack.lower()
            module_installed = reg.match(pack.replace("-","_"))
            if module_installed!=None and namemodule==module_installed.group(1):
                return module_installed.group(2)
    return False


#Retourne la liste des packages installés
@impmagic.loader(
    {'module':'subprocess'},
    {'module':'re', 'submodule': ['compile']},
    {'module':'sys_nxs.host', 'submodule': ['path_rep']},
    {'module':'template.regex', 'submodule': ['dependencies']}
)
def get_all_package(env_exe, refresh=False, clean_name=False):
    env_exe = env_exe.replace(path_rep[1], path_rep[0])
    stdout = []
    #Vérifie si le retour commande pip freeze existe déjà (pour éviter de l'appeler plusieurs fois)
    if refresh or not hasattr(__main__.nxs, 'already_installed'):
        cmd = [env_exe, '-m', 'pip', 'freeze']
        proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()
        __main__.nxs.already_installed = stdout
    else:
        stdout = __main__.nxs.already_installed
    
    stdout = stdout.decode('utf-8').split("\r\n")
    result = {}

    for pack in stdout:
        if len(pack):
            pack = pack.split("==")
            if len(pack)>1:
                if clean_name:
                    #result[pack[0].replace("-", "_").lower()] = pack[1]
                    result[pack[0].replace("-", "_")] = pack[1]
                else:
                    result[pack[0]] = pack[1]

    return result
    

#Récupérer la version de pip installé
@impmagic.loader(
    {'module':'subprocess'}
)
def get_pip_version(env_exe):
    cmd = [env_exe, '-m', 'pip', '--version']
    proc = subprocess.run(cmd, stdout=subprocess.PIPE)
    version = proc.stdout.decode().strip().split()[1]
    return version

#Installe la dernière version de pip si nécessaire
@impmagic.loader(
    {'module':'app.display', 'submodule': ['logs']},
    {'module':'package', 'submodule': ['package']}
)
def init_pip(env_exe):
    latest_pip = package.latest_version('pip')
    if latest_pip!=None:
        version_pip = get_pip_version(env_exe)
        if version_pip!=None:
            if not package.compare_version(latest_pip, "==", version_pip):
                logs("Nouvelle version de pip détectée", "warning")
                logs("Mise à jour de pip")
                upgrade_module(env_exe, "pip", force=True)
        else:
            logs("Pip n'est pas installé", "critical")
    else:
        logs("Impossible de trouver la version de pip", "critical")

#Parse des arguments envoyés
def arg_parse(string):
        array = []

        if len(string)>=1:
            arg = ""
            lock = None
            if isinstance(string, list):
                string = " ".join(string)
            for i,caracter in enumerate(string):
                if (caracter=="'" or caracter=='"') and (lock==None or caracter==lock):
                    if lock==None:
                        lock=caracter
                    else:
                        array.append(arg)
                        arg=""
                        lock=None
                else:
                    if caracter==" " and lock!=None:
                        arg+=caracter
                    elif caracter==" " and len(arg)>=1 and lock==None:
                        array.append(arg)
                        arg=""
                    elif caracter!=" ":
                        arg+=caracter
                        if i==len(string)-1:
                            array.append(arg)
                            arg=""
        return array


#Appel d'une commande dans un environnement
@impmagic.loader(
    {'module':'subprocess'},
    {'module': 'time', 'submodule': ['perf_counter']},
    {'module': 'datetime', 'submodule': ['timedelta', 'time']},
)
def command_shell(env_exe, command, args=None, timer=False):
    if isinstance(command, list):
        if command[0]=="python":
            command.pop(0)

        cmd = [env_exe] + command
    else:
        if command.startswith("python "):
            cmd = [env_exe, command.replace("python ","")]
        else:
            cmd = [env_exe, '-c', command]

    if args!=None and len(args)>0:
        if isinstance(args, str):
            cmd = cmd+arg_parse(args)
        if isinstance(args, list):
            cmd = cmd+args
    
    if timer:
        proc = subprocess.Popen(cmd, shell=True)
        
        st = perf_counter()
        proc.communicate()

        et = perf_counter()
        duree = timedelta(seconds=(et - st))
        print(f" Days          : {duree.days}\n Hours        : {duree.seconds//3600}\n Minutes      : {duree.seconds%3600//60}\n Seconds      : {duree.seconds%60}\n Milliseconds : {duree.microseconds//1000}\n Ticks        : {duree.microseconds*10}")
    else:
        proc = subprocess.Popen(cmd, shell=True)
        proc.communicate()
    '''    
    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    while True:
        output = proc.stdout.readline().decode("utf-8").strip()
        if output == "":
            break
        logs(output)

    for output in proc.stderr.readlines():
        logs(output.decode("utf-8").strip(), "error")
    '''

#Récupérer la version de python
@impmagic.loader(
    {'module':'subprocess'},
    {'module':'sys_nxs.host', 'submodule': ['path_rep']}
)
def get_py_version(env_exe):
    cmd = [env_exe.replace(path_rep[1], path_rep[0]), '-c', 'import platform; print(platform.python_version())']
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()

    if len(stdout)!=0:
        return stdout.decode().rstrip()
    else:
        return 0

#Récupérer le sys.path de python
@impmagic.loader(
    {'module':'subprocess'},
    {'module':'sys_nxs.host', 'submodule': ['path_rep']}
)
def get_py_syspath(env_exe):
    cmd = [env_exe.replace(path_rep[1], path_rep[0]), '-c', 'import sys; print ("\\n".join(sys.path));']
    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    if len(stdout)!=0:
        lines = stdout.decode('utf-8').splitlines()
        return list(filter(bool, lines))
    else:
        return []


#Vérifie si un module est installé
@impmagic.loader(
    {'module':'subprocess'},
    {'module':'chardet'},
    {'module':'app.display', 'submodule': ['logs']},
    {'module': 'template', 'submodule': ['default_conf']}
)
def installed_module(env_exe, namemodule, details=None, proxy=None, install=False):
    if proxy==None:
        proxy = __main__.nxs.conf.load(val='proxy', section='',default=default_conf.proxy)
        if proxy=="":
            proxy=None

    if namemodule!="pip" and namemodule!="setuptools":
        find = get_package(env_exe, namemodule)

        if not find:
            proc = subprocess.Popen([env_exe, '-c', f'"import pkg_resources; pkg_resources.get_distribution(\'{namemodule}\').version"'], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = proc.communicate()
            if not len(stderr) and len(stdout):
                return True

        if find!=False:
            find = find.replace("==","")
            if not details or find==details['version']:
                #logs(f"Module {namemodule} déjà installé dans la bonne version", "warning")
                return True
            else:
                if find in details['compatible']:
                    #logs(f"Module {namemodule} déjà installé dans une version compatible ({find})", "warning")
                    return True
                else:
                    logs(f"Module {namemodule} installé dans une version non compatible ({find})", "error")
                    if install==True:
                        logs("Mise à jour forcée du module", "warning")
                        res = upgrade_module(env_exe, namemodule, version=details['version'],force=True)
                        return res
                    else:
                        return False

    if install==True:
        logs(f"Installation du module {namemodule}")
        if details!=None:
            cmd = [env_exe, '-m', 'pip', 'install', namemodule+"=="+details['version']]
        else:
            cmd = [env_exe, '-m', 'pip', 'install', namemodule]
        if proxy!=None:
            cmd.append('--proxy='+proxy)

        full_log = __main__.nxs.conf.load(val='logs.full', section='',default=default_conf.logs_full)
        
        proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        while True:
            output = proc.stdout.readline().decode("utf-8").strip()
            if output == "":
                break
            if full_log:
                logs(output)


        #Pour éviter les erreurs sur un pip pas à jour
        d_error = False
        stderrr = proc.stderr.read()
        if stderrr!=None:
            if chardet.detect(stderrr)['encoding']!=None and "A new release of pip available" not in stderrr.decode(chardet.detect(stderrr)['encoding']).strip():
                proc.stderr.seek(0)
                for output in proc.stderr.readlines():
                    if len(output)>0 and d_error==False:
                        logs(f"ERREUR: Module {namemodule} non installé\nMessage d'erreur: {output.decode()}", "error")
                        d_error=True

                    if full_log:
                        logs(output.decode("utf-8").strip(), "error")

        if d_error==True:
            return False
        else:
            return True
    else:
        return False

#Installation de package en multithread si enable
@impmagic.loader(
    {'module':'concurrent.futures', 'as': 'worker'},
    {'module':'sys_nxs.host', 'submodule': ['path_rep']},
    {'module': 'template', 'submodule': ['default_conf']}
)
def install_pool(env_exe, installmodule, proxy=None, force=False):
    env_exe = env_exe.replace(path_rep[1], path_rep[0])
    if __main__.nxs.conf.load(val='threading', section='',default=default_conf.threading):
        threads = []

        with worker.ThreadPoolExecutor(max_workers=__main__.nxs.threadmax) as executor:
            futures = []
            for package in installmodule:
                details = None
                if isinstance(installmodule, dict):
                    details = installmodule[package]

                futures.append(executor.submit(install_module, env_exe, package, details, proxy, force))
            
            worker.wait(futures)

    else:
        for package in installmodule:
            install_module(env_exe,package, details=installmodule[package], proxy=proxy, force=force)


#Installation d'un module dans un environnement
@impmagic.loader(
    {'module':'subprocess'},
    {'module':'chardet'},
    {'module':'app.display', 'submodule': ['logs']},
    {'module': 'template', 'submodule': ['default_conf']}
)
def install_module(env_exe, namemodule, details=None, proxy=None, force=False):
    if proxy==None:
        proxy = __main__.nxs.conf.load(val='proxy', section='',default=default_conf.proxy)
        if proxy=="":
            proxy=None

    if namemodule!="pip" and namemodule!="setuptools":
        find = get_package(env_exe, namemodule)

        if find!=False:
            find = find.replace("==","")
            if details:
                if find==details['version']:
                    logs(f"Module {namemodule} déjà installé dans la bonne version", "warning")
                else:
                    if find in details['compatible']:
                        logs(f"Module {namemodule} déjà installé dans une version compatible ({find})", "warning")
                    else:
                        logs(f"Module {namemodule} installé dans une version non compatible ({find})", "error")

                    if force==True:
                        logs("Mise à jour forcée du module", "warning")
                        upgrade_module(env_exe, namemodule, version=details['version'],force=True)
            else:
                logs(f"Module {namemodule} déjà installé", "warning")

            return
    
    logs(f"Installation du module {namemodule}")
    if details!=None:
        cmd = [env_exe, '-m', 'pip', 'install', namemodule+"=="+details['version']]
    else:
        cmd = [env_exe, '-m', 'pip', 'install', namemodule]
    if proxy!=None:
        cmd.append('--proxy='+proxy)
    """
    cache_dir = pipcache("pip")
    if cache_dir!=None and os.path.exists(cache_dir):
        cmd.append('--cache-dir='+cache_dir)
    """

    full_log = __main__.nxs.conf.load(val='logs.full', section='',default=default_conf.logs_full)
    
    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    while True:
        output = proc.stdout.readline().decode("utf-8").strip()
        if output == "":
            break
        if full_log:
            logs(output)

    d_error = False
    #Pour éviter les erreurs sur un pip pas à jour
    stderrr = proc.stderr.read()
    if stderrr!=None:
        if chardet.detect(stderrr)['encoding']!=None and "A new release of pip available" not in stderrr.decode(chardet.detect(stderrr)['encoding']).strip():
            proc.stderr.seek(0)
            for output in proc.stderr.readlines():
                if len(output)>0 and d_error==False:
                    logs(f"ERREUR: Module {namemodule} non installé\nMessage d'erreur: {output.decode()}", "error")
                    d_error=True

                if full_log:
                    logs(output.decode("utf-8").strip(), "error")


#Suppression d'un module en multithread si enable
@impmagic.loader(
    {'module':'concurrent.futures', 'as': 'worker'},
    {'module':'sys_nxs.host', 'submodule': ['path_rep']},
    {'module': 'template', 'submodule': ['default_conf']}
)
def remove_pool(env_exe,namemodule):
    env_exe = env_exe.replace(path_rep[1], path_rep[0])
    if __main__.nxs.conf.load(val='threading', section='',default=default_conf.threading):
        threads = []

        with worker.ThreadPoolExecutor(max_workers=__main__.nxs.threadmax) as executor:
            futures = []
            for package in namemodule:
                futures.append(executor.submit(remove_module, env_exe, package))
            worker.wait(futures)

    else:
        for package in namemodule:
            remove_module(env_exe,package)

#Suppression d'un module dans un environnement
@impmagic.loader(
    {'module':'subprocess'},
    {'module':'app.display', 'submodule': ['logs']}
)
def remove_module(env_exe,namemodule):
    if get_package(env_exe, namemodule)==False:
        logs(f"Module {namemodule} non installé", "warning")
        return
    
    cmd = [env_exe, '-m', 'pip', 'uninstall', '-y', namemodule]
    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    if len(stdout)!=0:
        logs(f"Module {namemodule} supprimé")
    else:
        logs(f"ERREUR: Module {namemodule} non supprimé\nMessage d'erreur: {stderr.decode()}", "error")


#Installation de package en multithread si enable
@impmagic.loader(
    {'module':'concurrent.futures', 'as': 'worker'},
    {'module':'sys_nxs.host', 'submodule': ['path_rep']},
    {'module': 'template', 'submodule': ['default_conf']}
)
def upgrade_pool(env_exe, installmodule, proxy=None, force=False):
    env_exe = env_exe.replace(path_rep[1], path_rep[0])
    if __main__.nxs.conf.load(val='threading', section='',default=default_conf.threading):
        threads = []

        with worker.ThreadPoolExecutor(max_workers=__main__.nxs.threadmax) as executor:
            futures = []
            for package in installmodule:
                details = None
                if isinstance(installmodule, dict):
                    details = installmodule[package]
                    futures.append(executor.submit(upgrade_module, env_exe, package, details['version'], proxy, force))

            worker.wait(futures)

    else:
        for package in installmodule:
            upgrade_module(env_exe,package, installmodule[package]['version'], proxy=proxy, force=force)
    
#Upgrade d'un module dans un environnement
@impmagic.loader(
    {'module':'subprocess'},
    {'module':'chardet'},
    {'module':'app.display', 'submodule': ['logs']},
    {'module':'sys_nxs.host', 'submodule': ['path_rep']},
    {'module': 'template', 'submodule': ['default_conf']}
)
def upgrade_module(env_exe, namemodule, version=None, proxy=None, force=False, reinstall=False):
    logs(f"Mise à jour du module {namemodule}")

    env_exe = env_exe.replace(path_rep[1], path_rep[0])

    if version!=None:
        for_install = namemodule+"=="+version
    else:
        for_install = namemodule

    find = get_package(env_exe, namemodule)
    if find!=False:
        find = find.replace("==","")
        if version!=None and find==version:
            logs(f"Module {namemodule} déjà installé dans la bonne version")
            return True

    if proxy==None:
        proxy = __main__.nxs.conf.load(val='proxy', section='',default=default_conf.proxy)
        if proxy=="":
            proxy=None

    if namemodule!="pip" and namemodule!="setuptools":
        if get_package(env_exe, namemodule)==False:
            logs(f"Module {namemodule} non installé", "warning")
            return

    cmd = [env_exe, '-m', 'pip', 'install', for_install, '--upgrade']
    if proxy!=None:
        cmd.append('--proxy='+proxy)

    if force==True:
        cmd.append('--force')
    
    if reinstall==True:
        cmd.append('--force-reinstall')

    full_log = __main__.nxs.conf.load(val='logs.full', section='',default=default_conf.logs_full)

    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    while True:
        output = proc.stdout.readline().decode("utf-8").strip()
        if output == "":
            break
        if full_log:
            logs(output)

    d_error = False
    #Pour éviter les erreurs sur un pip pas à jour
    stderrr = proc.stderr.read()
    if stderrr!=None:
        if chardet.detect(stderrr)['encoding']!=None and "A new release of pip available" not in stderrr.decode(chardet.detect(stderrr)['encoding']).strip():
            proc.stderr.seek(0)
            for output in proc.stderr.readlines():
                if len(output)>0 and d_error==False:
                    logs(f"ERREUR: Module {namemodule} non mis à jour\nMessage d'erreur: {proc.stderr.readlines()}", "error")
                    d_error=True

                if full_log:
                    logs(output.decode("utf-8").strip(), "error")

    if d_error==False:
        logs(f"Module {namemodule} mis à jour")
        return True

    return False

#Créer un environnement dans le cache
@impmagic.loader(
    {'module':'__main__'},
    {'module':'uuid', 'submodule': ['uuid1']},
    {'module': 'template', 'submodule': ['default_conf']},
    {'module':'os.path', 'submodule': ['expanduser','join']}
)
def create_temp_env(installmodule=None, upgradepip=True):
    name = str(uuid1())
    cachedir = join(expanduser(__main__.nxs.conf.load(val='cache-dir', section='',default=default_conf.cache_dir)), name)
    __main__.ToDoClear.append(cachedir)
    create_environment(cachedir, name=name, installmodule=None, upgradepip=upgradepip, prompt="temp-"+name)

    return name, cachedir


#Création d'un environnement
@impmagic.loader(
    {'module':'sys_nxs.host', 'submodule': ['path_rep']},
    {'module':'zpp_config', 'submodule': ['Config']},
    {'module':'os.path', 'submodule': ['exists', 'isdir']},
    {'module':'os'},
    {'module':'virtualenv'},
    {'module':'app.display', 'submodule': ['logs']},
    {'module':'sys_nxs.host', 'submodule': ['path_reg']},
    {'module': 'template', 'submodule': ['default_conf']},
    {'module':'sys'}
)
def create_environment(virtdir, name=None, installmodule=None, clear=False, upgradepip=True, symlinks=False, sitepackages=False, proxy=None, prompt=None):
    if proxy==None:
        proxy = __main__.nxs.conf.load(val='proxy', section='',default=default_conf.proxy)
        if proxy=="":
            proxy=None
    if name!=None:
        logs(f"Environnement {name}")
    else:
        logs(f"Environnement {virtdir}")

    logs("Création du contexte")
    virtdir = path_reg(virtdir)
    context = Context(virtdir, name, prompt=prompt)

    args = []
    args.append("--no-download")
    args.append("--no-periodic-update")
    #args.append("--no-setuptools")
    #args.append("--no-pip")
    #args.append("--no-wheel")

    if not symlinks:
        args.append("--always-copy")
    
    if sitepackages:
        args.append("--system-site-packages")

    if clear:
        args.append("--clear")

    if prompt is not None:
        args.extend(["--prompt", prompt])

    #args.append("--python")
    #args.append("PATHPYTHON")
    args.append(context.env_dir)

    if clear==False and exists(virtdir) and isdir(virtdir) and len(os.listdir(virtdir))!=0:
        logs("Le dossier n'est pas vide", "warning")
        sys.exit()
    
    logs("Création de l'environnement")
    try:
        virtualenv.cli_run(args)
    except Exception as err:
        logs(err, "error")

    if not exists(context.bin_path) or not exists(context.lib_path):
        logs("ERREUR: Le dossier de l'environnement n'a pas été créé", "error")
        sys.exit()
    
    if not exists(context.cfg_path):
        logs("ERREUR: Le fichier de config n'a pas été créé", "error")
        sys.exit()
    
    if not exists(context.env_exe):
        logs("ERREUR: L'exécutable n'a pas été copié", "error")
        sys.exit()
    
    if not exists(context.bin_path+path_rep[0]+"activate"):
        logs("ERREUR: Les scripts d'activation n'ont pas été créé", "error")
        sys.exit()

    if os.name=='nt':
        pipname = 'pip.exe'
    else:
        pipname = 'pip'
    if not exists(context.bin_path+path_rep[0]+pipname):
        logs("ERREUR: Pip n'a pas été installé", "error")
        sys.exit()

    if upgradepip:
        logs("Recherche de mise à jour")
        upgrade_module(context.env_exe,"pip", proxy=proxy)
        upgrade_module(context.env_exe,"setuptools", proxy=proxy)

    if installmodule!=None:
        if isinstance(installmodule, str):
            installmodule = installmodule.split(",")

        install_pool(context.env_exe, installmodule, None, proxy)

    logs("Environnement créé")

    venv_data = {}
    venv_data['env_dir'] = context.env_dir 
    venv_data['env_name'] = context.env_name 
    venv_data['env_exe'] = context.env_exe
    venv_data['bin_path'] = context.bin_path
    c = Config(context.cfg_path)
    data = c.load(val='version', section='',default="N.A")
    venv_data['version'] = data

    return venv_data

#Suppression d'un environnement
@impmagic.loader(
    {'module':'os.path', 'submodule': ['exists']},
    {'module':'shutil', 'submodule': ['rmtree']},
    {'module':'app.display', 'submodule': ['logs']}
)
def remove_environment(virtdir):
    if exists(virtdir):
        logs("Suppresion d'un dossier d'environnement")
        try:
            rmtree(virtdir)
        except PermissionError:
            logs("ERREUR: Autorisation refusée pour supprimer le dossier d'environnement", "error")
        except Exception as err:
            logs(f"Error: {err}", "error")
    else:
        logs("Le dossier d'environnement n'existe pas", "error")

#Ouvrir l'environnement
@impmagic.loader(
    {'module':'sys_nxs.host', 'submodule': ['path_rep']},
    {'module':'os.path', 'submodule': ['exists', 'isdir']},
    {'module':'os', 'submodule': ['name', 'environ']},
    {'module':'subprocess'},
    {'module':'pexpect'},
    {'module':'app.display', 'submodule': ['logs']},
    {'module':'sys_nxs.host', 'submodule': ['path_reg', 'path_rep', 'get_os']},
    {'module':'sys'}
)
def open_environment(virtdir, projectfolder="", shell=False):
    context = Context(path_reg(virtdir))

    if exists(context.bin_path) and isdir(context.bin_path):
        if shell:
            cmd = context.env_exe
        else:
            OSType = get_os()
            if OSType=='cmd' or OSType=='cmd.exe':
                activate_file='activate.bat'
                if exists(context.bin_path+path_rep[0]+activate_file):
                    cmd = ['cmd', '/k']
                    target = ""
                    ### Add alias CMD
                    if projectfolder!="":
                        target+=f"doskey cdproject=cd {projectfolder} & "
                    else:
                        target+=f"doskey cdproject=echo "+"Le dossier de projet n'est pas configuré"+" & "
                    target+=f"doskey shellenv={context.env_exe} & "
                    target+=f'doskey help=echo "'+"cdproject  se déplacer dans le dossier du projet`nshellenv   accéder au shell python`ndeactivate    sortir de l'environnement"+'" & '
                    ### END Add alias
                    target+="pushd "+context.bin_path.replace(path_rep[1], path_rep[0])+" & .\\"+activate_file+" & popd"
                    cmd.append(target)
                else:
                    logs("ERREUR: Le script d'activation n'existe pas", "error")
                    sys.exit()

            elif OSType=='powershell.exe' or name=='nt':
                activate_file='activate.ps1'
                if exists(context.bin_path+path_rep[0]+activate_file):
                    cmd = ['powershell', '-NoExit', '-Command']
                    target = ""
                    ### Add alias PS1
                    """
                    """
                    if projectfolder!="":
                        target+='function cdproject{cd "'+projectfolder+'"}; '
                    else:
                        target+='function cdproject{Write-Host '+"Le dossier de projet n`'est pas configuré"+'}; '
                    target+='function shellenv{'+context.env_exe+'}; '
                    target+='function help{Write-Host "'+"cdproject  se déplacer dans le dossier du projet`nshellenv   accéder au shell python`ndeactivate    sortir de l'environnement"+'"}; '
                    ### END Add alias
                    target+='. "'+context.bin_path+path_rep[0]+activate_file+'"'

                    cmd.append(target)
                else:
                    logs("ERREUR: Le script d'activation n'existe pas", "error")
                    sys.exit()
            
            else:
                activate_file='activate'
                if exists(context.bin_path+path_rep[0]+activate_file):

                    cmd = []
                    target = ""
                    ### Add alias Linux
                    """
                    if projectfolder!="":
                        target+=f"alias cdproject=\'cd {projectfolder}\' && "
                    else:
                        target+="alias cdproject=\'echo "+"Le dossier de projet n'est pas configuré"+"\' && "
                    target+=f"alias shellenv='{context.env_exe}' && "
                    target+=f'alias help=\'echo -e "'+"cdproject  se déplacer dans le dossier du projet\nshellenv   accéder au shell python\ndeactivate    sortir de l'environnement"+'"\' && '
                    ### END alias Linux
                    target+=". "+context.bin_path+path_rep[0]+activate_file
                    cmd.append(target)
                    """

                    target=". "+context.bin_path+path_rep[0]+activate_file

                    shell = environ.get('SHELL')
                    interact = pexpect.spawn(shell, ['-i'])
                    interact.sendline(target)
                    interact.interact(escape_character=None)

                    interact.close()
                    return
                else:
                    logs("ERREUR: Le script d'activation n'existe pas", "error")
                    sys.exit()

        subprocess.call(cmd, shell=True)
    else:
        logs("ERREUR: Dossier d'environnement introuvable", "error")


@impmagic.loader(
    {'module':'__main__'},
    {'module':'shutil'},
    {'module':'app.display', 'submodule': ['logs']}, 
    {'module':'structure.check', 'submodule': ['get_root_project']}, 
    {'module':'template.toml_format', 'submodule': ['DEFAULT_TOML']},
    {'module': 'template', 'submodule': ['default_conf']},
    {'module': 'os', 'submodule': ['chdir']},
    {'module': 'sys_nxs.host', 'submodule': ['path_rep']},
    {'module':'os.path', 'submodule': ['exists', 'isabs', 'join', 'split', 'expanduser', 'dirname']}
)
def clone_environment(env, dest_env, tfile, sandbox=False):
    if exists(join(get_root_project(), env['env_exe'])):
        chdir(get_root_project())

        if exists(dest_env):
            logs("Le répertoire de destination existe déjà", "warning")
            return False, {}
        env_folder = __main__.nxs.conf.load(val='virtualenvs.foldername', section='',default=default_conf.virtualenvs_foldername)
        cachedir = expanduser(__main__.nxs.conf.load(val='cache-dir', section='',default=default_conf.cache_dir)).replace(path_rep[1], path_rep[0])
        env_dir = split(env['env_dir'])[1]
        logs(f"Clonage en cours de l'environnement {env['env_name']}")

        environment_dir = env['env_dir'].replace(path_rep[1], path_rep[0])

        if isabs(environment_dir) and environment_dir.startswith(cachedir):
            appname = tfile.get_key(DEFAULT_TOML['projectname']['name'],DEFAULT_TOML['projectname']['section'])

            parent = dirname(join(cachedir,env_folder,appname,env_dir))
            source = join(cachedir,env_folder,appname,env_dir)
        else:
            parent = dirname(environment_dir)
            source = environment_dir

        dest = join(parent, dest_env)
        shutil.copytree(source, dest)

        venv_data = {}
        venv_data['env_dir'] = dest.replace(path_rep[0], path_rep[1])
        venv_data['env_name'] = dest_env
        venv_data['env_exe'] = env['env_exe'].replace(env['env_dir'], dest).replace(path_rep[0], path_rep[1])
        venv_data['bin_path'] = env['bin_path'].replace(env['env_dir'], dest).replace(path_rep[0], path_rep[1])
        venv_data['version'] = env['version']

        
        fixup_scripts(source, dest, get_py_version(env['env_exe']), venv_data['bin_path'])
        fix_symlink_if_necessary(source, dest)

        #Reinstallation de pip pour le link sur le nouvel environnement
        pipversion = get_pip_version(venv_data['env_exe'])
        upgrade_module(venv_data['env_exe'], 'pip', version=pipversion, reinstall=True)


        logs("Environnement cloné")
        return True, venv_data
    else:
        logs("L'environnement est introuvable", "critical")

    return False, {}

#Fonction récursive pour faire la liste des packages installés dans l'env en fonction des dépendances du fichier toml
#ATTENTION NE PREND EN COMPTE QUE LES PATH DE PROFONDEUR 1
@impmagic.loader(
    {'module':'os'}
)
def get_package_installed(cmd, deps, lockable, path=None):
    if path!=None and path not in lockable:
        lockable[path]={}

    for package in deps:
        if package!="windows" and package!="linux":
            a = get_package(cmd, package)
            if a==False:
                logs(f"{package} non installé", "warning")
            else:
                version = a.replace("==","")

                if path!=None:
                    lockable[path][package]=version
                else:
                    lockable[package]=version

        elif package=="windows" and os.name=="nt":
            lockable = get_package_installed(cmd, deps['windows'], lockable, "windows")
        elif package=="linux" and os.name!="nt": 
            lockable = get_package_installed(cmd, deps['linux'], lockable, "linux")
    return lockable


#Filtrer le dictionnaire de dépendances à partir d'une liste 
@impmagic.loader(
    {'module':'tomlkit'}
)
def filter_deps(filtre, deps):
    for package, value in deps.copy().items():
        if package=='windows' or package=='linux':
            res = filter_deps(filtre, value)
            if len(res):
                deps[package] = res
            
            if not len(deps[package]):
                del deps[package]
        elif package not in filtre:
            del deps[package]

    return deps

@impmagic.loader(
    {'module':'app.display', 'submodule': ['logs']}
)
def change_default(tfile, env_name):
    last = None
    for env in tfile.doc['venv']:
        name = env
        env = tfile.doc['venv'][env]
        if 'default' in env and env['default']==True:
            last = name

    for env in tfile.doc['venv']:
        name = env
        env = tfile.doc['venv'][env]
        if env_name==name:
            tfile.edit_key('default', True, "venv."+name)
            if last!=None:
                tfile.edit_key('default', False, "venv."+last)
            logs("Environnement par défaut modifié")


## Fixing function from virtualenv-clone
## https://github.com/edwardgeorge/virtualenv-clone
def _dirmatch(path, matchwith):
    matchlen = len(matchwith)
    if (path.startswith(matchwith)
        and path[matchlen:matchlen + 1] in [os.sep, '']):
        return True
    return False

@impmagic.loader(
    {'module':'itertools'},
    {'module':'os', 'submodule': ['walk', 'remove', 'symlink']},
    {'module':'os.path', 'submodule': ['join', 'islink', 'realpath']}
)
def fix_symlink_if_necessary(src_dir, dst_dir):
    for dirpath, dirnames, filenames in walk(dst_dir):
        for a_file in itertools.chain(filenames, dirnames):
            full_file_path = join(dirpath, a_file)
            if islink(full_file_path):
                target = realpath(full_file_path)
                if target.startswith(src_dir):
                    new_target = target.replace(src_dir, dst_dir)
                    remove(full_file_path)
                    symlink(new_target, full_file_path)


@impmagic.loader(
    {'module':'itertools'},
    {'module':'re'},
    {'module':'os', 'submodule': ['walk']},
    {'module':'os.path', 'submodule': ['join', 'islink', 'isfile']}
)
def fixup_scripts(old_dir, new_dir, version, bin_dir):
    #bin_dir = join(new_dir, env_bin_dir)
    root, dirs, files = next(walk(bin_dir))
    pybinre = re.compile(r'pythonw?([0-9]+(\.[0-9]+(\.[0-9]+)?)?)?$')

    for file_ in files:
        filename = join(root, file_)
        #Changement le chemin du venv dans les scripts activate
        if file_ == 'activate' or file_.startswith('activate.'):

            with open(filename, 'rb') as f:
                data = f.read().decode('utf-8')

            data = data.replace(old_dir, new_dir)
            with open(filename, 'wb') as f:
                f.write(data.encode('utf-8'))

        #Relink des liens symboliques
        elif islink(filename):
            fixup_link(filename, old_dir, new_dir)

        #Remplacement des shebang
        elif isfile(filename):
            fixup_script_(root, file_, old_dir, new_dir, version)


@impmagic.loader(
    {'module':'os.path', 'submodule': ['join', 'normcase', 'abspath']}
)
def fixup_script_(root, file_, old_dir, new_dir, version):
    old_shebang = '#!%s/bin/python' % normcase(abspath(old_dir))
    new_shebang = '#!%s/bin/python' % normcase(abspath(new_dir))

    filename = join(root, file_)
    with open(filename, 'rb') as f:
        if f.read(2) != b'#!':
            # no shebang
            return
        f.seek(0)
        lines = f.readlines()

    if not lines:
        return

    def rewrite_shebang(version=None):
        shebang = new_shebang
        if version:
            shebang = shebang + version
        shebang = (shebang + '\n').encode('utf-8')
        with open(filename, 'wb') as f:
            f.write(shebang)
            f.writelines(lines[1:])

    try:
        bang = lines[0].decode('utf-8').strip()
    except UnicodeDecodeError:
        return

    short_version = bang[len(old_shebang):]

    if not bang.startswith('#!'):
        return
    elif bang == old_shebang:
        rewrite_shebang()
    elif (bang.startswith(old_shebang) and bang[len(old_shebang):] == version):
        rewrite_shebang(version)
    elif (bang.startswith(old_shebang) and short_version and bang[len(old_shebang):] == short_version):
        rewrite_shebang(short_version)
    else:
        return


@impmagic.loader(
    {'module':'os', 'submodule': ['readlink', 'symlink', 'rename']},
    {'module':'os.path', 'submodule': ['join', 'dirname', 'abspath', 'isabs']}
)
def fixup_link(filename, old_dir, new_dir, target=None):
    if target is None:
        target = readlink(filename)

    origdir = dirname(abspath(filename)).replace(new_dir, old_dir)
    if not isabs(target):
        target = abspath(join(origdir, target))
        rellink = True
    else:
        rellink = False

    if _dirmatch(target, old_dir):
        if rellink:
            target = target[len(origdir):].lstrip(os.sep)
        else:
            target = target.replace(old_dir, new_dir, 1)

    tmpfn = "%s.new" % filename
    symlink(target, tmpfn)
    rename(tmpfn, filename)



import __main__
import impmagic

@impmagic.loader(
    {'module':'requests'},
    {'module':'zpp_config', 'submodule':['Config']},
    {'module': 'template', 'submodule': ['default_conf']}
)
def get(url, params=None, headers=None):
    proxy = __main__.nxs.conf.load(val='proxy', section='',default=default_conf.proxy)
    if proxy!=None and len(proxy):
        proxies = {
            "http": proxy,
            "https": proxy
        }

    else:
        proxies = None

    response = requests.get(url, proxies=proxies, params=params, headers=headers)

    return response


import impmagic

@impmagic.loader(
    {'module':'sys_nxs.connect', 'submodule':['get']}
)
def list_all_licenses():
    response = get("https://api.github.com/licenses")
    result = {}

    if response.status_code==200:
        licenses = response.json()
        for license in licenses:
            response = get(license['url'])
            if response.status_code==200:
                licenses_content = response.json()
                result[license['key']] = {}
                result[license['key']]['key'] = license['key']
                result[license['key']]['name'] = license['name']
                result[license['key']]['license'] = licenses_content['body']
    return result

@impmagic.loader(
    {'module':'__main__'},
    {'module':'os.path', 'submodule':['exists']},
    {'module':'json', 'submodule':['load', 'dump']}
)
def get_data_licenses():
    if exists(__main__.nxs.license_cache_file):
        with open(__main__.nxs.license_cache_file) as json_file:
            licenses = load(json_file)
    else:
        licenses = list_all_licenses()
        with open(__main__.nxs.license_cache_file, 'w') as json_file:
            dump(licenses, json_file)
    return licenses

@impmagic.loader(
    {'module':'Levenshtein'}
)
def detect_license(file):
    with open(file) as f:
        content = f.read()

    licenses = get_data_licenses()

    for license in licenses:
        license = licenses[license]
        distance = Levenshtein.distance(content, license['license'])
        similarity = 1 - (distance / max(len(content), len(license['license'])))
        if (similarity * 100)>95:
            return license['name']

    return False


def licenses_availables(name):
    licenses = get_data_licenses()

    for license in licenses:
        license = licenses[license]
        if name==license['name'] or name==license['key']:
            return True
    return False


@impmagic.loader(
    {'module':'app.display', 'submodule':['print_nxs', 'prompt']}
)
def get_license_content(name):
    licenses = get_data_licenses()

    if sum(1 for lic in licenses.keys() if name in lic)>1:
        options = [lic for lic in licenses.keys() if name in lic]
        choice = "0"
        while not choice.isdigit() or int(choice)<=0 or int(choice) >len(options):
            print_nxs("Plusieurs licences disponibles. Veuillez choisir la bonne:")
            for i, name in enumerate(options):
                print_nxs(f"{i+1}. {name}")
            choice = prompt("Choix: ")

        name = options[int(choice)-1]

    for license in licenses:
        license = licenses[license]
        if name==license['name'] or name==license['key']:
            return license['license']
    return False


def get_license_fullname(name):
    licenses = get_data_licenses()

    for license in licenses:
        license = licenses[license]
        if name==license['name'] or name==license['key']:
            return license['name']
    
    return False

@impmagic.loader(
    {'module':'glob', 'submodule':['glob']}
)
def license_exist():
    for file in glob('*'):
        if file.startswith('LICENSE'):
            return file
    return False


import impmagic

class Repo:
	@impmagic.loader(
		{'module':'pygit2', 'submodule': ['Repository', 'Signature', 'init_repository']},
		{'module':'os.path', 'submodule': ['abspath']}
	)
	def __init__(self, repo_path, branch_name=None, create=True):
		self.repo_path = abspath(repo_path)

		try:
			self.repo = Repository(self.repo_path)
		except:
			if create:
				self.repo = init_repository(self.repo_path)
			else:
				return

		try:
			self.author = self.repo.default_signature
			self.committer = self.repo.default_signature
		except:
			self.author = Signature("nxs_default", "nxs_default@example.com")
			self.committer = Signature("nxs_default", "nxs_default@example.com")

		#Si nouveau projet, initialiser la branche master
		if create and branch_name==None:
			self.init_master()

		if branch_name==None:
			branch_name = self.get_active_branch()

		if branch_name not in self.list_branch():
			if not len(list(self.repo.references)):
				self.init_master()
			
			if branch_name!="master":
				self.add_branch(branch_name)
		
		self.branch_name = branch_name

	def commit_exist(self, commit_id):
		try:
			self.repo.revparse_single(commit_id)
			return True
		except:
			return False

	#Liste les branches disponibles
	def list_branch(self, local=False, remote=False):
		if (not local and not remote) or (local and remote):
			return list(self.repo.branches)

		if local and not remote:
			return list(self.repo.branches.local)

		if remote and not local:
			return list(self.repo.branches.remote)

	#Récupérer la liste des branches avec les informations
	@impmagic.loader(
		{'module':'pygit2'},
		{'module':'datetime', 'submodule': ['datetime']}
	)
	def get_branch_info(self, local=False, remote=False):
		infos = {}
		active_branch = self.get_active_branch()

		for branch_name in self.list_branch(local, remote):
			branch = self.repo.branches[branch_name]
			commit = branch.peel(pygit2.Commit)

			if active_branch==branch_name:
				active = True
			else:
				active = False

			infos[branch_name] = {
				'last_commit_id': commit.hex,
				'last_commit_message': commit.message,
				'last_commit_author': f'{commit.author.name} <{commit.author.email}>',
				'last_commit_committer': f'{commit.committer.name} <{commit.committer.email}>',
				'last_commit_date': str(datetime.fromtimestamp(commit.commit_time)),
				'parents': commit.parents,
				'parent_ids': commit.parent_ids,
				'active_branch': active,
			}

		return infos

	#Récupérer la liste des commits
	@impmagic.loader(
		{'module':'pygit2'},
		{'module':'datetime', 'submodule': ['datetime']}
	)
	def get_commit_info(self):
		infos = {}

		# Parcourir tous les commits du dépôt
		for commit in self.repo.walk(self.repo.head.target, pygit2.GIT_SORT_TIME | pygit2.GIT_SORT_TOPOLOGICAL):
			# Obtenir les références associées au commit
			commit_refs = []
			for ref_name in self.repo.listall_references():
				ref = self.repo.lookup_reference(ref_name)
				if ref.target.hex == commit.id.hex:
					commit_refs.append(ref_name)
			commit_refs = ', '.join(commit_refs)


			# Afficher les informations du commit
			commit_refs = commit_refs.replace('refs/heads/', '')

			infos[commit.hex] = {
				'commit_id': commit.id,
				'commit_hash': str(commit.id)[:7],
				'commit_message': commit.message,
				'commit_message_resume': commit.message.strip().split('\n')[0],
				'commit_author': f'{commit.author.name} <{commit.author.email}>',
				'commit_committer': f'{commit.committer.name} <{commit.committer.email}>',
				'commit_time': str(datetime.fromtimestamp(commit.commit_time)),
				'commit_refs': commit_refs,
				'parent_ids': commit.parent_ids,
			}

		return infos

	@impmagic.loader(
		{'module':'pygit2'},
		{'module':'app.display', 'submodule': ['print_nxs']}
	)
	def log(self, oneline=False):
		commit = self.repo.head.peel(pygit2.Commit)

		for commit, info in self.get_commit_info().items():
			if info['commit_id'] == self.repo.head.target:
				info['commit_refs'] = "HEAD -> "+info['commit_refs']

			if oneline:
				if len(info['commit_refs']):
					print(f"{info['commit_hash']} ({info['commit_refs']}) {info['commit_message_resume']}")
				else:
					print(f"{info['commit_hash']} {info['commit_message_resume']}")
			else:
				if len(info['commit_refs']):
					print(f"\ncommit: {info['commit_id']} ({info['commit_refs']})")
				else:
					print(f"\ncommit: {info['commit_id']}")

				if len(info['parent_ids']):
					print(f"Parents: {', '.join([str(element)[:7] for element in info['parent_ids']])}")

				print(f"Author: {info['commit_author']}")

				if info['commit_author']!=info['commit_committer']:
					print(f"Committer: {info['commit_committer']}")

				print(f"Date: {info['commit_time']}")

				print(f"Message: {info['commit_message']}")


	#Récupérer le nom de la branche active
	def get_active_branch(self):
		return self.repo.head.shorthand

	#Création d'un ref
	def add_ref(self, branch_name):
		tree = self.repo.TreeBuilder().write()

		commit_oid = self.repo.create_commit(None, self.author, self.committer, "Commit initial", tree, [])
		branch_ref = self.repo.create_reference(self.get_ref(branch_name), commit_oid)

	#Renvoie la référence de branche
	def get_ref(self, branch_name):
		return f"refs/heads/{branch_name}"

	#Création de la première branche
	def init_master(self):
		tree = self.repo.TreeBuilder().write()

		# Créer le commit initial avec l'arbre vide
		commit_oid = self.repo.create_commit('HEAD', self.author, self.committer, "Commit initial", tree, [])


	#Ajoute une branche au repo (à partir du master, d'une autre branche ou d'un commit spécifique)
	@impmagic.loader(
		{'module':'pygit2'},
		{'module':'app.display', 'submodule': ['logs']}
	)
	def add_branch(self, branch_name, source_branch=None, commit_id='HEAD'):
		if source_branch!=None and not self.get_active_branch()==source_branch:
			status_code = self.switch_branch(source_branch)
			if not status_code:
				return

		if branch_name in self.list_branch():
			logs(f"La branche {branch_name} existe déjà", "error")
			return False

		if not self.commit_exist(commit_id):
			logs("Commit ID invalid", "error")
			exit()

		try:
			commit = self.repo.revparse_single(commit_id).peel(pygit2.Commit)
			self.repo.create_branch(branch_name, commit)
			return True
		except Exception as err:
			logs(err, "error")
			return False

	#Supprimer une branche au repo
	@impmagic.loader(
		{'module':'app.display', 'submodule': ['logs']}
	)
	def delete_branch(self, branch_name):
		self.check_branch(branch_name)
		try:
			self.repo.branches.delete(branch_name)
			return True
		except Exception as err:
			logs(err, "error")
			return False

	#Switch sur une nouvelle branche
	@impmagic.loader(
		{'module':'app.display', 'submodule': ['logs']}
	)
	def switch_branch(self, branch_name, create=False):
		if branch_name in self.repo.branches:
			self.repo.checkout(self.get_ref(branch_name))
			self.branch_name = branch_name
			return True
		elif branch_name not in self.repo.branches:
			if create:
				logs(f"Création de la branche {branch_name}", "info")
				self.add_branch(branch_name)
				self.repo.checkout(self.get_ref(branch_name))
				self.branch_name = branch_name
				return True
			else:
				logs(f"La branche {branch_name} n'existe pas", "error")
				return False

	#Renommer une branche
	@impmagic.loader(
		{'module':'app.display', 'submodule': ['logs']}
	)
	def rename_branch(self, branch_name, new_branch_name):
		self.check_branch(branch_name)

		if new_branch_name in self.list_branch():
			logs(f"La branche {new_branch_name} existe déjà", "error")

		try:
			ref = self.repo.lookup_reference(self.get_ref(branch_name))
			ref.rename(self.get_ref(new_branch_name))
			return True
		except Exception as err:
			logs(err, "error")
			return False

	#Fusionner deux branches
	@impmagic.loader(
		{'module':'pygit2'},
		{'module':'app.display', 'submodule': ['logs']}
	)
	def merge_branch(self, source_branch, target_branch=None):
		try:
			self.check_branch(source_branch)

			if not target_branch:
				target_branch = self.get_active_branch()

			if not self.get_active_branch()==target_branch:
				status_code = self.switch_branch(target_branch)
				if not status_code:
					return

			source_ref = self.repo.branches[source_branch]

			source_commit = source_ref.peel(pygit2.Commit)

			index = self.repo.merge(source_commit.id)

			# Créer un nouvel objet de commit avec l'index fusionné
			signature = self.author
			merge_commit_id = self.repo.create_commit(
				self.get_ref(target_branch), signature, signature,
				"Merge: {} into {}".format(source_branch, target_branch),
				self.repo.index.write_tree(), [self.repo.head.target, source_commit.id]
			)
			# Mettre à jour la branche cible pour pointer vers le nouveau commit de fusion
			#target_ref.set_target(merge_commit_id)
			return True
		except Exception as err:
			logs(err, "error")
			return False


	#Retourne le hash de l'objet
	@impmagic.loader(
		{'module':'os.path', 'submodule': ['exists', 'isfile']}
	)
	def hash_object(self, file_path):
		if exists(file_path) and isfile(file_path):
			with open(file_path, 'rb') as file:
				file_content = file.read()

				# Créer un objet blob à partir du contenu du fichier
				blob_oid = self.repo.create_blob(file_content)

				return str(blob_oid)

	#Vérifie si la branche existe, sinon quitte
	@impmagic.loader(
		{'module':'app.display', 'submodule': ['logs']}
	)
	def check_branch(self, branch_name):
		if branch_name not in self.list_branch():
			logs(f"La branche {branch_name} n'existe pas", "error")
			exit()

	#Afficher le statut du repo
	@impmagic.loader(
		{'module':'pygit2'},
		{'module':'app.display', 'submodule': ['print_nxs']}
	)
	def status(self):
		for name, branch in self.get_branch_info().items():
			if branch['active_branch']:
				print(f"Active branch: {name}")
				print(f"Last commit:")
				print(f"   Id:{branch['last_commit_id']}")
				if len(branch['parent_ids']):
					print(f"   Parents: {', '.join([str(element)[:7] for element in branch['parent_ids']])}")
				print(f"   Date: {branch['last_commit_date']}")
				print(f"   Author: {branch['last_commit_author']}")
				if branch['last_commit_author']!=branch['last_commit_committer']:
					print(f"   Committer: {branch['last_commit_committer']}")
				print(f"   Message: {branch['last_commit_message']}")


		status_file = {'modified': [], 'deleted': [], 'new': [], 'conflicted': [], 'not_commited': []}
		# Afficher les fichiers modifiés dans l'espace de travail
		status = self.repo.status()
		for file in status:
			if status[file] == pygit2.GIT_STATUS_WT_MODIFIED:
				status_file['modified'].append(file)

			elif status[file] == pygit2.GIT_STATUS_WT_DELETED:
				status_file['deleted'].append(file)

			elif status[file] == pygit2.GIT_STATUS_WT_NEW:
				status_file['new'].append(file)

			elif status[file] == pygit2.GIT_STATUS_CONFLICTED:
				status_file['conflicted'].append(file)

			elif status[file] == pygit2.GIT_STATUS_INDEX_NEW or status[file] == pygit2.GIT_STATUS_INDEX_MODIFIED:
				status_file['not_commited'].append(file)


		# Afficher les fichiers modifiés dans l'espace de travail
		if len(status_file['modified']):
			file = '\n   '.join(status_file['modified'])
			print(f"\nChanges not staged for commit:\n   {file}")

		# Afficher les fichiers supprimés dans l'espace de travail
		if len(status_file['deleted']):
			file = '\n   '.join(status_file['deleted'])
			print(f"\nDeleted files:\n   {file}")

		# Afficher les fichiers non suivis
		if len(status_file['new']):
			file = '\n   '.join(status_file['new'])
			print(f"\nUntracked files:\n   {file}")

		# Afficher les fichiers en conflit
		if len(status_file['conflicted']):
			file = '\n   '.join(status_file['conflicted'])
			print(f"\nConflicted files:\n   {file}")

		# Afficher les fichiers ajoutés à l'index
		if len(status_file['not_commited']):
			file = '\n   '.join(status_file['not_commited'])
			print(f"\nChanges to be committed:\n   {file}")



	#Ajoute à l'index l'ensemble des modifications et commit
	def git_add_and_commit(self, commit_message="", branch_name=None):
		if not branch_name:
			branch_name = self.branch_name

		self.check_branch(branch_name)

		try:
			# Ajouter tous les modifications à l'index
			self.repo.index.add_all()
			#Ecrit l'index
			self.repo.index.write()
		except Exception as err:
			logs(f"Indexation impossible: \n{err}", "error")
			return None


		#Lecture de l'arbre de l'index
		tree = self.repo.index.write_tree()

		#Obtention de l'objet ref de la branche
		branch_ref = self.repo.lookup_reference(self.get_ref(branch_name))

		#Obtention de l'objet commit de la branche
		branch_commit = self.repo[branch_ref.target]

		try:
			#Création du commit
			commit_oid = self.repo.create_commit(self.get_ref(branch_name), self.author, self.committer, commit_message, tree, [branch_commit.id])
		except Exception as err:
			logs(f"Commit impossible: \n{err}", "error")
			return None

		# Récupérer l'identifiant du nouveau commit
		commit = self.repo.get(commit_oid)
		
		return commit.hex

	#Restauration depuis un commit
	@impmagic.loader(
		{'module':'pygit2'},
		{'module':'app.display', 'submodule': ['logs']}
	)
	def reset(self, commit_id, mode="hard"):
		if not self.commit_exist(commit_id):
			logs("Commit ID invalid", "error")
			exit()

		mode_list = {"soft": pygit2.GIT_RESET_SOFT, "hard": pygit2.GIT_RESET_HARD, "mixed": pygit2.GIT_RESET_MIXED}
		if mode in mode_list:
			try:
				self.repo.reset(commit_id, mode_list[mode])
			except Exception as err:
				logs(err, "error")
		else:
			print("Mode invalid")


	#Restaurer un fichier spécifique
	@impmagic.loader(
		{'module':'os', 'submodule': ['getcwd', 'chdir']},
		{'module':'os.path', 'submodule': ['relpath', 'abspath', 'split']},
		{'module':'app.display', 'submodule': ['logs']}
	)
	def restore(self, file, dest=None, commit_id='HEAD'):
		if not self.commit_exist(commit_id):
			logs("Commit ID invalid", "error")
			exit()

		init = None
		if self.repo_path not in getcwd():
			logs("You are outside the repo folder. Moving to the root of the repo", "info")
			init = getcwd()
			chdir(self.repo_path)

		file_path = relpath(abspath(file), self.repo_path)

		# Obtenir le contenu du fichier à partir du commit
		commit = self.repo.revparse_single(commit_id)

		file_array = list(split(file))
		#Suppression du champ vide si on cherche pas dans un dossier
		if "" in file_array:
			file_array.remove('')
		blob = self.search_file_from_commit_tree(commit.tree, file_array)
		
		if init and dest:
			chdir(init)
			file_out = abspath(dest)
		elif dest:
			file_out = abspath(dest)
		else:
			file_out = file_path

		#Si on veut restaurer un dossier entier
		if isinstance(blob, pygit2.Tree):
			self.restore_folder(blob, file_out)
		else:
			if blob!=None:
				# Récupérer le contenu du blob
				content = self.repo[blob.id].data.decode('utf-8')
			else:
				logs(f"File {file_path} not found", "error")
				return False

			try:
				# Écrire le contenu du fichier dans l'arborescence de travail
				with open(file_out, 'w') as file:
					file.write(content)
			except Exception as err:
				logs(err, "error")
				return False

			return True

	#Recherche un fichier à travers l'arborescence d'un commit
	@impmagic.loader(
		{'module':'pygit2'}
	)
	def search_file_from_commit_tree(self, commit_tree, file_array):
		for entry in commit_tree:
			if entry.name==file_array[0]:
				file_array.pop(0)
				if len(file_array):
					if isinstance(entry, pygit2.Tree):
						return self.search_file_from_commit_tree(entry, file_array)
				else:
					return self.repo[entry.id]
		return None

	#Restauration d'un dossier git complet
	@impmagic.loader(
		{'module':'os', 'submodule': ['mkdir']},
		{'module':'os.path', 'submodule': ['join', 'exists', 'isdir']}
	)
	def restore_folder(self, commit_tree, dest):
		if not exists(dest) or not dest:
			mkdir(dest)

		for entry in commit_tree:
			if isinstance(entry, pygit2.Tree):
				self.restore_folder(entry, join(dest, entry.name))
			else:
				content = self.repo[entry.id].data.decode('utf-8')

				try:
					# Écrire le contenu du fichier dans l'arborescence de travail
					with open(join(dest, entry.name), 'w') as file:
						file.write(content)
				except Exception as err:
					logs(err, "error")
					return False
		return None

	@impmagic.loader(
		{'module':'os', 'submodule': ['getcwd', 'chdir']},
		{'module':'os.path', 'submodule': ['relpath', 'abspath']}
	)
	def tree(self, commit_id='HEAD'):
		if not self.commit_exist(commit_id):
			logs("Commit ID invalid", "error")
			exit()

		init = None
		if self.repo_path not in getcwd():
			logs("You are outside the repo folder. Moving to the root of the repo", "info")
			init = getcwd()
			chdir(self.repo_path)

		# Obtenir le contenu du fichier à partir du commit
		commit = self.repo.revparse_single(commit_id)

		self.tree_display(commit.tree)

	@impmagic.loader(
		{'module':'pygit2'}
	)
	def tree_display(self, commit_tree, lvl=0, arbo_display=[]):
		for i, entry in enumerate(commit_tree):
			if isinstance(entry, pygit2.Tree):
				print_pattern(i, len(commit_tree), arbo_display)
				print(entry.name)
				if i==len(commit_tree)-1:
					new_arbo = arbo_display+['   ']
				else:
					new_arbo = arbo_display+['│  ']

				self.tree_display(entry, lvl+1, arbo_display=new_arbo)
			else:
				print_pattern(entry.name, len(commit_tree), arbo_display)
				print(entry.name)

def print_pattern(index, max, arbo):
	arbo = "".join(arbo)
	if index==max-1:
		print(f"{arbo}└─", end="")
	else:
		print(f"{arbo}├─", end="")